from enum import Enum
from abc import ABC, abstractmethod

import pika
import time
from uuid import uuid4

import logging
from pokeman.composite.config.configs import BasicConfig

LOGGER = logging.getLogger(__name__)


class ConnectionType(Enum):
    URL = {'type': 'URL', 'format': 'amqp://<user>:<password>@<host>:<port>'}
    HOST = {'type': 'HOST', 'format': "<resource-dns>"}


class ChannelStatus(Enum):
    INACTIVE = 0x0
    ACTIVE = 0x1
    CLOSED = 0x2


class AbstractConnectionParameters(ABC):
    """
    Abstract connection parameters class
    """

    def __init__(self, connstr, method=ConnectionType.URL, username='guest', password='guest', config=BasicConfig()):
        """
        Abstract initialization method for the connection object.

        :param connstr: The provided connection string.
        :type connstr: str

        :param method: The provided connection method.
        :type method: ConnectionType

        :param username: The provided AMQP broker username.
        :type username: str

        :param password: The provided AMQP broker password.
        :type password: str

        :param config: The provided connection configuration object.
        :type config: BasicConfig
        """
        self.connection_string = connstr
        self.method = method
        self.username = username
        self.password = password
        self.config = config

    def _append_url_parameters(self, uri):
        """
        Appends the connection configuration parameters to the connection
        url.

        :param uri: The string for connecting to AMQP broker.
        :type uri: str

        :return: The connection string with parameters.
        :rtype: str

        .. note::
            the uri should look something like this:
            *amqp://guest:guest@localhost:5672*
        """
        parameters = '?'
        config = {
            'connection_attempts': self.config.CONNECTION_ATTEMPTS,
            'heartbeat': self.config.HEARTBEAT,
            'retry_delay': self.config.RETRY_DELAY
        }
        for param, value in config.items():
            if parameters != '?':
                parameters += '&'
            parameters += ('{PARAM}={VALUE}'.format(PARAM=param,
                                                    VALUE=value))
        connection_url = uri + parameters
        return connection_url

    def resolve_uri_parameters(self, uri):
        """
        This method resolves the uri parameters from the provided connection
        string.

        :param uri: The provided AMQP broker connection string.
        :type uri: str

        :return: The resolved connection uri.
        :rtype: str
        """
        LOGGER.debug('Resolving {CONNECTION} uri parameters'.format(CONNECTION=self.__class__.__name__))
        resolved_uri = self._append_url_parameters(uri)
        LOGGER.debug('Resolving {CONNECTION} uri parameters OK!'.format(CONNECTION=self.__class__.__name__))
        return resolved_uri


class AbstractConnection(ABC):
    """
    Abstract connection composite structure.
    """

    def __init__(self, parameters):
        """
        Abstract initialization method for the connection object.

        :param parameters: The provided connection parameters.
        :type parameters: ConnectionParameters

        """
        self.cp = parameters
        self.connection = None
        self.channels = {}

    def set_channel_status(self, channel_id, status):
        """
        This method sets the channel status.

        :param channel_id: The channel ID
        :type channel_id: str

        :param status: The channel status
        :type status: ChannelStatus
        """
        self.channels[channel_id]['status'] = status

    def open_channel(self):
        """
        This method opens a new channel for the connection.

        :return: The channel ID
        :rtype: str
        """
        channel_id = str(uuid4())
        self.channels[channel_id]['channel'] = self.connection.channel()
        self.set_channel_status(channel_id=channel_id, status=ChannelStatus.INACTIVE)
        return channel_id

    def close_channel(self, channel_id):
        """
        This method closes a new channel for the connection.
        """
        self.channels[channel_id]['channel'].close()
        self.set_channel_status(channel_id=channel_id, status=ChannelStatus.CLOSED)

    def close_all_channels(self):
        """
        This method closes all the open channels for the connection.
        """
        for channel, context in self.channels.items():
            if context['status'] == ChannelStatus.INACTIVE or context['status'] == ChannelStatus.ACTIVE:
                context['channel'].close()

    @abstractmethod
    def connect(self, **kwargs):
        """
        Abstract connect method.
        """
        pass  # pragma:

    @abstractmethod
    def disconnect(self):
        """
        Abstract disconnect method.
        """
        pass


class ConnectionParameters(AbstractConnectionParameters):
    pass


class Connection(AbstractConnection):
    """
    The Blocking composite connection with the AMQP broker.
    """

    def connect(self):
        """
        This method connects to AMQP broker, and sets the object connection variable.

        :return: The pika connection handle.
        :rtype: pika.BlockingConnection

        .. note::
            The connection parameters are set depending on the connection type:
                * pika.URLParameters(url) when the ConnectionType == URL, useful
                  when connecting to remote AMQP server.
                * pika.ConnectionParameters(host) when the ConnectionType == HOST,
                  useful when connecting to Docker container in same cluster.

        .. note::
            The while loop assures that the producer will fail to start due tot an
            AMQPConnectionError that is the result of AMQP broker still in starting up
            phase.
        """
        LOGGER.debug('Pokeman connecting to AMQP broker')
        parameters = None
        connected = False
        if self.cp.method == ConnectionType.URL:
            parameters = pika.URLParameters(
                url=self.cp.resolve_uri_parameters(uri=self.cp.connection_string)
            )
        elif self.cp.method == ConnectionType.HOST:
            parameters = pika.ConnectionParameters(
                host=self.cp.connection_string,
                credentials=pika.credentials.PlainCredentials(
                    username=self.cp.username,
                    password=self.cp.password,
                ),
                connection_attempts=self.cp.config.CONNECTION_ATTEMPTS,
                heartbeat=self.cp.config.HEARTBEAT,
                retry_delay=self.cp.config.RETRY_DELAY
            )
        retries = 0
        _connection = None
        while not connected and retries < 6:
            time.sleep(2)
            retries += 1
            try:
                _connection = pika.BlockingConnection(parameters=parameters)
            # TODO: Too broad exception handling here, zoom-in
            except AttributeError:
                raise AttributeError('The connection object has no attribute close')
            else:
                LOGGER.debug('Pokeman connecting to AMQP broker OK!')
                self.connection = _connection
        # TODO MAYBE: _connection.add_on_connection_blocked_callback(self.on_connection_blocked)
        #   def on_connection_blocked(self, _unused_connection, _unused_fram):
        #       """
        #       This method will be called if the blocked_connection_timeout limit is overdue.
        #       :return:
        #       """
        #       pass

    def disconnect(self):
        """
        This method disconnects the connection with the AMQP broker.
        """
        if self.connection is not None:
            LOGGER.debug('Closing Pokeman connection')
            try:
                self.connection.close()
            except Exception:
                LOGGER.exception('Closing Pokeman connection FAILED!')
                raise ConnectionError('Closing Pokeman connection FAILED!')
            else:
                if self.connection is not None and not self.connection.is_closed:
                    raise ConnectionError('Closing Pokeman connection FAILED!')
                LOGGER.debug('Closing Pokeman connection OK!')


class SelectConnection(AbstractConnection):
    """
    The Selecting composite connection with the AMQP broker.
    """
    def connect(self, on_open_callback, on_open_error, on_close_callback):
        """
        This method connects to AMQP broker, and sets the object connection variable.

        .. note::
            The connection parameters are set depending on the connection type:
                * pika.URLParameters(url) when the ConnectionType == URL, useful
                  when connecting to remote AMQP server.
                * pika.ConnectionParameters(host) when the ConnectionType == HOST,
                  useful when connecting to Docker container in same cluster.

        .. note::
            The while loop assures that the producer will fail to start due tot an
            AMQPConnectionError that is the result of AMQP broker still in starting up
            phase.
        """
        LOGGER.debug('Pokeman connecting to AMQP broker')
        parameters = None
        connected = False
        if self.cp.method == ConnectionType.URL:
            parameters = pika.URLParameters(
                url=self.cp.resolve_uri_parameters(uri=self.cp.connection_string)
            )
        elif self.cp.method == ConnectionType.HOST:
            parameters = pika.ConnectionParameters(
                host=self.cp.connection_string,
                credentials=pika.credentials.PlainCredentials(
                    username=self.cp.username,
                    password=self.cp.password,
                ),
                connection_attempts=self.cp.config.CONNECTION_ATTEMPTS,
                heartbeat=self.cp.config.HEARTBEAT,
                retry_delay=self.cp.config.RETRY_DELAY
            )
        retries = 0
        _connection = None
        while not connected and retries < 6:
            time.sleep(2)
            retries += 1
            try:
                _connection = pika.SelectConnection(
                    parameters=parameters,
                    on_open_callback=on_open_callback,
                    on_open_error_callback=on_open_error,
                    on_close_callback=on_close_callback
                )
            # TODO: Too broad exception handling here, zoom-in
            except AttributeError:
                raise AttributeError('The connection object has no attribute close')
            else:
                LOGGER.debug('Pokeman connecting to AMQP broker OK!')
                self.connection = _connection

    def disconnect(self):
        """
        This method disconnects the connection with the AMQP broker.
        """
        if self.connection is not None and not self.connection.is_closing and not self.connection.is_closed:
            LOGGER.debug('Closing Pokeman connection')
            try:
                self.connection.close()
            except Exception:
                LOGGER.exception('Closing Pokeman connection FAILED!')
                raise ConnectionError('Closing Pokeman connection FAILED!')
            else:
                if self.connection is not None and not self.connection.is_closed:
                    raise ConnectionError('Closing Pokeman connection FAILED!')
                LOGGER.debug('Closing Pokeman connection OK!')



# # docker run -it --rm --name rabbitmq -p 5672:5672 -p 15672:15672 rabbitmq:3-management
#
# def nonex():
#     pass
#
# # class TestUnits(unittest.TestCase):
# def test_async():
#     import pokeman as pm
#
#     async_config = pokeman.composite.config.AsyncConsumerTopicExchangeConfig(topic="search_engine", event="request")
#     async_consumer = pm.ReconnectingAsyncConsumer(connstr='amqp://guest:guest@localhost:5672',
#                                                   config=async_config)
#     async_consumer.run(func=nonex)
#
# test_async()
#
# # TODO: Bij verlies verbinding async consumer werkt de boel niet meer, en blijven messages hangen.


from pokeman.coatings.message_constructions._abc_message import AbstractBasicMessageConstruction
from pokeman.amqp_resources.globals import Exchange, RoutingKey

import logging


LOGGER = logging.getLogger(__name__)


# TODO: Check topic exchange configuration
class BasicMessage(AbstractBasicMessageConstruction):
    """
    Basic Message Enterprise Integration Pattern for message 
    construction > Basic Message.

    A Sentinel is declared default for various pattern parameters,
    as the possible None declaration is a posteriori.
    """
    _SENTINEL = Ellipsis

    def __init__(
            self,
            app_id=_SENTINEL,
            exchange=_SENTINEL,
            routing_key=_SENTINEL
    ):
        """
        :param app_id: The provided app_id
        :type app_id: str

        :param exchange: The provided exchange.
        :type exchange: Exchange

        :param routing_key: The provided routing key.
        :type routing_key: pokeman.RoutingKey
        """
        self._exchange = exchange
        self._routing_key = routing_key
        self._exchange_converter = lambda _pkid: Exchange(
            specific_poker=self._specific_poker_setter(
                specific_poker=_pkid,
                lambda_call=True
            )).exchange_name if self._exchange is self._SENTINEL else self._exchange.exchange_name
        self._routing_key_converter = lambda: RoutingKey().key if self._routing_key is self._SENTINEL else self._routing_key.key
        self.exchange = self._exchange_setter(exchange=self._exchange_converter)
        self.routing_key = self._routing_key_setter(routing_key=self._routing_key_converter())
        self.app_id = self._app_id_setter(app_id=app_id)

    def _app_id_setter(self, app_id):
        if app_id is self._SENTINEL:
            return "pokeman_app"
        else:
            return app_id

    def _exchange_setter(self, exchange):
        return exchange

    def _routing_key_setter(self, routing_key):
        if routing_key is None:
            raise ValueError('Please provide a routing key')
        return routing_key

    # TODO: Add Fanout and Headers exchange types

    def _specific_poker_setter(self, specific_poker, lambda_call=False):
        if specific_poker is self._SENTINEL:
            return None
        elif lambda_call is False:
            return specific_poker.poker_id
        elif lambda_call is True:
            return specific_poker
        else:
            raise ValueError('Unresolvable specific poker parameter')


import json

import logging

LOGGER = logging.getLogger(__name__)


def start(self):
    self.start_consuming()


def on_message(self, channel, method, properties, body):
    """
    Invoked by pika when a message is delivered from the AMQP broker. The
    channel is passed for convenience. The basic_deliver object that
    is passed in carries the exchange, routing key, delivery tag and
    a redelivered flag for the message. The properties passed in is an
    instance of BasicProperties with the message properties and the body
    is the message that was sent.

    :param channel: The channel object.
    :type channel: pika.channel.Channel

    :param method: basic_deliver method.
    :type method: pika.Spec.Basic.Deliver

    :param properties: The properties.
    :type properties: pika.Spec.BasicProperties

    :param body: The message body.
    :type body: bytes
    """
    try:
        print('message received')
        print(properties.correlation_id)
        if properties.correlation_id == self.correlation_id_reference:
            print("SUCCEEDEEDRT")
            self.callback_method(json.loads(body), properties)
            self.acknowledge_message(method.delivery_tag)
            self.channel.stop_consuming()
    except Exception:
        LOGGER.exception("Synchronous callback method exception:")


from pokeman.utils.custom_abc import ABCMeta, abstract_attribute


# TODO: Add more arguments https://pika.readthedocs.io/en/stable/modules/channel.html#pika.channel.Channel.basic_consume
class AbstractBasicMessagingEndpoint(metaclass=ABCMeta):
    """
    Abstract base class for Enterprise Integration Patterns,
    in specific Messaging Endpoints.
    """

    @abstract_attribute
    def exchange(self):
        pass

    @abstract_attribute
    def queue(self):
        pass

    @abstract_attribute
    def callback_method(self):
        pass

    @abstract_attribute
    def qos(self):
        pass



import sys
import os
import contextlib
import inspect
import logging
import unittest

LOGGER = logging.getLogger(__name__)


class _Notset:
    def __repr__(self):
        return "<notset>"


notset = _Notset()


class TestAttributes:
    @classmethod
    @contextlib.contextmanager
    def patch(cls, target, name, value=notset):
        """
        Factory method for temporarily monkeypatching objects.

        This method is created because the _pytest.monkeypatch.
        MonkeyPatch() object causes fixture leaking between various
        tests and between files.
        """
        _setattr = []
        if value is notset:
            if not isinstance(value, str):
                raise TypeError(
                    "Use patchfactory(target, name, value) or "
                    "setattr(target, value) with target being a dotted "
                    "import string"
                )
            value = name
            # name, target = derive_importpath(target, raising)

        pre_patched_value = getattr(target, name, notset)
        if pre_patched_value is notset:
            raise AttributeError("{TARGET} has no attribute {NAME}".format(TARGET=target, NAME=name))

        # avoid class descriptors like staticmethod/classmethod
        if inspect.isclass(target):
            pre_patched_value = target.__dict__.get(name, notset)
        _setattr.append((target, name, pre_patched_value))
        setattr(target, name, value)

        try:
            LOGGER.debug("Patch yield for: {TARGET}.{NAME} OK!".format(TARGET=target.__dict__.get(name, notset),
                                                                  NAME=name))
            yield target
        finally:
            for obj, name, value in reversed(_setattr):
                if value is not notset:
                    setattr(obj, name, value)
                else:
                    delattr(obj, name)
            LOGGER.debug("Patch yield for: {TARGET}.{NAME} DESTROYED!".format(TARGET=target.__dict__.get(name, notset),
                                                                       NAME=name))
            _setattr[:] = []

    @classmethod
    @contextlib.contextmanager
    def chdir(cls, path):
        """
        Change the current working directory to the specified path.
        Path can be a string or a py.path.local object.
        """
        pre_scenario_cwd = os.getcwd()
        if hasattr(path, "chdir"):
            path.chdir()
        else:
            os.chdir(path)
        try:
            yield
        finally:
            os.chdir(pre_scenario_cwd)



from abc import ABC, abstractmethod
from uuid import uuid4

# Message Constructions
from .message_constructions import BasicMessage
from.messaging_endpoints import PollingConsumer, BasicMessageConsumer, SelectiveConsumer
from .resolver_attribute_methods.selective_consumer import start, on_message

import logging

LOGGER = logging.getLogger(__name__)

# TODO: Methods get_app_id and set_channel to abstract Resolvers instead of Resolver


class AbstractMessageConstructionResolver(ABC):
    """
    Abstract base clase for a EIP Message Construction resolver.
    """
    def __init__(self, coating):
        self.coating = coating
        self.blueprint = {
            'type': None,
            'app_id': None,
        }
        self.resolve_coating()

    def resolve_coating(self):
        self.get_build_type()
        self.get_app_id()
        self.set_exchange()
        self.set_routing_key()
        self.set_response_queue()

    def get_build_type(self):
        if 'Resolver' not in self.__class__.__name__:
            raise NameError('Resolver not found in {CLASS_NAME}'.format(CLASS_NAME=self.__class__.__name__))
        self.blueprint['type'] = self.__class__.__name__.replace('Resolver', '')

    @abstractmethod
    def get_app_id(self):
        pass

    @abstractmethod
    def set_exchange(self):
        pass

    @abstractmethod
    def set_routing_key(self):
        pass

    def set_response_queue(self):
        pass


class AbstractMessagingEndpointResolver(ABC):
    """
    Abstract base clase for a EIP Messaging Endpoint resolver.
    """
    def __init__(self, coating):
        self.coating = coating
        self.blueprint = {
            'type': None,
            'app_id': None
        }
        self.resolve_coating()

    def resolve_coating(self):
        self.get_build_type()
        self.set_exchange()
        self.set_queue()
        self.set_callback_method()
        self.set_qos()

    def get_build_type(self):
        if 'Resolver' not in self.__class__.__name__:
            raise NameError('Resolver not found in {CLASS_NAME}'.format(CLASS_NAME=self.__class__.__name__))
        self.blueprint['type'] = self.__class__.__name__.replace('Resolver', '')

    @abstractmethod
    def set_exchange(self):
        pass

    @abstractmethod
    def set_queue(self):
        pass

    @abstractmethod
    def set_callback_method(self):
        pass

    @abstractmethod
    def set_qos(self):
        pass

    def set_response_queue(self):
        pass


class BasicMessageResolver(AbstractMessageConstructionResolver):
    """
    Template resolver for EIP Message Construction > BasicMessage
    """
    def get_app_id(self):
        """
        This method sets the app id in the blueprint.
        """
        self.blueprint['app_id'] = self.coating.app_id

    def set_exchange(self):
        """
        This methods sets the exchange in the blueprint.
        """
        self.blueprint['exchange'] = self.coating.exchange

    def set_routing_key(self):
        """
        This method sets the routing key in the blueprint
        """
        self.blueprint['routing_key'] = self.coating.routing_key


class PollingConsumerResolver(AbstractMessagingEndpointResolver):
    """
    Template resolver for EIP Messaging Endpoint > PollingConsumer
    """

    def set_exchange(self):
        """
        This methods sets the exchange in the blueprint.
        """
        self.blueprint['exchange'] = self.coating.exchange

    def set_queue(self):
        self.blueprint['queue'] = self.coating.queue

    def set_callback_method(self):
        self.blueprint['callback_method'] = self.coating.callback_method

    def set_qos(self):
        self.blueprint['qos'] = self.coating.qos


class BasicMessageConsumerResolver(AbstractMessagingEndpointResolver):
    """
    Template resolver for EIP Messaging Endpoint > BasicMessageConsumer
    """

    def set_exchange(self):
        """
        This methods sets the exchange in the blueprint.
        """
        self.blueprint['exchange'] = self.coating.exchange

    def set_queue(self):
        self.blueprint['queue'] = self.coating.queue

    def set_callback_method(self):
        self.blueprint['callback_method'] = self.coating.callback_method

    def set_qos(self):
        self.blueprint['qos'] = self.coating.qos


class SelectiveConsumerResolver(BasicMessageConsumerResolver):
    """
    Template resolver for EIP Messaging Endpoint > SelectiveConsumer
    """

    def __init__(self, coating):
        super().__init__(coating=coating)
        self.set_reference()

    def set_reference(self):
        self.blueprint['reference'] = {
            'correlation_id': {
                'value': self.coating.correlation_id,
                'start_method': start,
                'on_message_method': on_message
            }
        }


# TODO MAYBE: Maybe bring the Ptypes validation to the Wingman.
class Wingman:
    """
    The Wingman is considered the coating resolver servant of the Foreman,
    it takes a provided EIP coating object and makes a blueprint
    of it that can be digested by the builders that the Foreman
    manages.
    """
    # Message Constructions
    _templates = {
        BasicMessage: BasicMessageResolver,
        PollingConsumer: PollingConsumerResolver,
        BasicMessageConsumer: BasicMessageConsumerResolver,
        SelectiveConsumer: SelectiveConsumerResolver
    }

    def __init__(self):
        """
        This method resets the Wingman on initialization.
        """
        self.reset()

    def reset(self):
        """
        This method resets the blueprint object to None.
        """
        self._blueprint = None

    def resolve(self, coating):
        """
        This method resolves the provided coating object to a bluerint.

        :param coating: The provided EIP coating.
        :type coating: pokeman.coating

        :return: The blueprint.
        """
        if coating.__class__ in self._templates.keys():
            _coating = coating
        else:
            raise ValueError("No coating resolver implemented for {COATING}".format(COATING=coating.__class__.__name__))

        resolver = self._templates[_coating.__class__](coating=_coating)
        self._blueprint = resolver.blueprint
        blueprint = self._blueprint
        self.reset()
        return blueprint



from abc import ABC, abstractmethod
from uuid import uuid4
import functools
import time
import os
import pika
from pika.exceptions import AMQPConnectionError, UnroutableError
import json

from pokeman.utils.custom_abc import abstract_attribute
from pokeman.coatings.resolvers import Wingman
from pokeman.coatings.ptypes import Ptypes
from . import __all__ as _c__all__

import logging

LOGGER = logging.getLogger(__name__)


class AbstractSynchronousProducerBuilder(ABC):
    """
    Abstract base class for a synchronous producer builder.
    """

    @abstract_attribute
    def producer(self):
        pass

    @abstractmethod
    def digest_blueprint(self):
        pass

    @abstractmethod
    def open_channel(self):
        pass

    @abstractmethod
    def apply_delivery_confirmations(self):
        pass

    @abstractmethod
    def setup_exchange(self):
        pass

    @abstractmethod
    def setup_routing_key(self):
        pass


class AbstractAsynchronousConsumerBuilder(ABC):
    """
    Abstract base class for a asynchronous consumer builder.
    """

    @abstract_attribute
    def consumer(self):
        pass

    @abstractmethod
    def digest_blueprint(self):
        pass

    @abstractmethod
    def setup_connection(self):
        pass

    @abstractmethod
    def setup_exchange(self):
        pass

    @abstractmethod
    def setup_queue(self):
        pass

    @abstractmethod
    def setup_callback_method(self):
        pass

    @abstractmethod
    def setup_qos(self):
        pass


class AbstractSynchronousConsumerBuilder(ABC):
    """
    Abstract base class for a synchronous consumer builder.
    """

    @abstract_attribute
    def consumer(self):
        pass

    @abstractmethod
    def digest_blueprint(self):
        pass

    @abstractmethod
    def open_channel(self):
        pass

    @abstractmethod
    def setup_exchange(self):
        pass

    @abstractmethod
    def setup_queue(self):
        pass

    @abstractmethod
    def setup_callback_method(self):
        pass

    @abstractmethod
    def setup_qos(self):
        pass

    @abstractmethod
    def elevate_setup(self):
        pass


class SynchronousProducerBuilder(AbstractSynchronousProducerBuilder):

    def __init__(self, connection, blueprint):
        """
        This method initializes the producer builder and sets the
        connection and the by the Foreman already on compatibility
        checked build template.

        :param connection: The provided AMQP broker connection that is attached
        to the Pokeman.
        :type connection: pokeman.Pokeman.connection

        :param blueprint: The resolved template from the Wingman.
        :type blueprint: pokeman.coating._resolvers.Wingman._blueprint
        """
        self.reset()
        self.connection = connection
        self.blueprint = blueprint

    def reset(self):
        """
        The production builder is reset on initialization and after a
        delivery of a producer.
        """
        self._producer = SynchronousProducer()

    @property
    def producer(self):
        """
        When the producer delivery is invoked, the builder resets itself for
        a new producing task.

        :return: The built producer object.
        :rtype: pokeman.coating._builders.SynchronousProducer
        """
        producer = self._producer
        self.reset()
        return producer

    def digest_blueprint(self):
        """
        This method envokes blueprint digestion by the builder.
        It detects if a multistep build process is necessary and
        sets the channel build context.
        """
        LOGGER.debug('{BUILDER} digesting blueprint'.format(BUILDER=self.__class__.__name__))
        LOGGER.debug('{BUILDER} digesting blueprint OK!'.format(BUILDER=self.__class__.__name__))

    def start_independent_build(self):
        """
        This method invokes the standard building process for
        the producer builder. The Foreman can invoke custom builds by
        instructing the builder on it's own.
        """
        self.open_channel()
        self.apply_delivery_confirmations()  # TODO: Make choice in composite
        self.setup_exchange()
        self.setup_routing_key()

    def set_app_id(self):
        """
        This method sets the app id for the producer.
        """
        self._producer.APP_ID = self.blueprint['app_id']

    def open_channel(self):
        """
        This method will open a new channel with AMQP broker and attach
        it to the producer.
        """
        LOGGER.debug('Opening a new channel')
        new_channel = self.connection.channel()
        self._producer.channel = new_channel
        LOGGER.debug('Opening a new channel OK!')

    def apply_delivery_confirmations(self):
        """
        This method will apply delivery confirmations on a BlockingConnection.

        .. note::
            When the message is not received correctly a pika.exceptions.
            UnroutableError will be raised.
        """
        LOGGER.debug('Applying delivery confirmations')
        self._producer.channel.confirm_delivery()
        LOGGER.debug('Applying delivery confirmations OK!')

    def setup_exchange(self):
        """
        Set up the exchange on the AMQP broker by declaring the exchange
        on the current context channel.
        """
        _exchange_name = self.blueprint['exchange']
        LOGGER.debug('Binding exchange {EXCHANGE_NAME} to {APP_ID} producer'.format(
            EXCHANGE_NAME=_exchange_name,
            APP_ID=self.blueprint['app_id']
        )
        )
        self._producer.exchange = _exchange_name
        LOGGER.debug('Binding exchange {EXCHANGE_NAME} to {APP_ID} producer OK!'.format(
            EXCHANGE_NAME=_exchange_name,
            APP_ID=self.blueprint['app_id']
        )
        )

    def setup_routing_key(self):
        self._producer.routing_key = self.blueprint['routing_key']


class SynchronousProducer:
    """
    The synchronous producer unfinished intangible.
    A builder Ptypes allowed needs to configure and
    deliver this object.
    """
    APP_ID = None
    CONTENT_TYPE = 'application/json'
    CONTENT_ENCODING = None
    HEADERS = None
    DELIVERY_MODE = 2
    PRIORITY = None
    CORRELATION_ID = None
    REPLY_TO = None
    EXPIRATION = None
    MESSAGE_ID = None
    TIMESTAMP = None
    TYPE = None
    USER_ID = None
    CLUSTER_ID = None

    def __init__(self):
        self.channel = None
        self.exchange = None
        self.routing_key = None
        self._message_number = 0

    def publish(self, message,
                content_type=CONTENT_TYPE,
                content_encoding=CONTENT_ENCODING,
                headers=HEADERS,
                delivery_mode=DELIVERY_MODE,
                priority=PRIORITY,
                correlation_id=CORRELATION_ID,
                reply_to=REPLY_TO,
                expiration=EXPIRATION,
                message_id=MESSAGE_ID,
                timestamp=TIMESTAMP,
                type=TYPE,
                user_id=USER_ID,
                cluster_id=CLUSTER_ID
                ):
        """
        This method publishes a message to AMQP broker, incrementing
        the number of messages variable.

        :param message: The message
        :type message: str|unicode|dict|list|int|float

        :param headers: The headers, unicode key,values are allowed
        :type headers: dict

        .. note::
            The first if statement checks if the channel is not dead or
            already in use (open).
        """

        if self.channel is None or not self.channel.is_open:
            raise ConnectionError('This producer has no active AMQP broker connection')

        if not correlation_id:
            correlation_id = str(uuid4())

        properties = pika.BasicProperties(
            content_type=content_type,
            content_encoding=content_encoding,
            headers=headers,
            delivery_mode=delivery_mode,
            priority=priority,
            correlation_id=correlation_id,
            reply_to=reply_to,
            expiration=expiration,
            message_id=message_id,
            timestamp=timestamp,
            type=type,
            user_id=user_id,
            app_id=self.APP_ID,
            cluster_id=cluster_id
        )
        try:
            self.channel.basic_publish(
                exchange=self.exchange,
                routing_key=self.routing_key,
                body=json.dumps(message, ensure_ascii=False),
                properties=properties
            )
        except UnroutableError as e:
            """There was nog delivery confirmation from the broker"""
            LOGGER.exception(e)
            raise
        except AMQPConnectionError as e:
            """The connection could not be established"""
            LOGGER.exception(e)
            raise
        except Exception as e:
            """Any other exception"""
            LOGGER.exception(e)
            raise
        self._message_number += 1
        LOGGER.debug("Published message # {MESSAGE_NUMBER}".format(MESSAGE_NUMBER=self._message_number))
        return {
            "correlation_id": correlation_id
        }


class AsynchronousConsumerBuilder(AbstractAsynchronousConsumerBuilder):

    def __init__(self, connection, blueprint):
        """
        This method initializes the consumer builder and sets the
        connection and the by the Foreman already on compatibility
        checked build template.

        :param connection: The provided AMQP broker connection that is attached
        to the Pokeman.
        :type connection: pokeman.Pokeman.connection

        :param blueprint: The resolved template from the Wingman.
        :type blueprint: pokeman.coating._resolvers.Wingman._blueprint
        """
        self.reset()
        self.connection = connection
        self.blueprint = blueprint

    def reset(self):
        """
        The production builder is reset on initialization and after a
        delivery of a consumer.
        """
        self._consumer = AsynchronousConsumer()

    @property
    def consumer(self):
        """
        When the consumer delivery is invoked, the builder resets itself for
        a new producing task.

        :return: The built consumer object.
        :rtype: pokeman.coating._builders.SynchronousConsumer
        """
        consumer = self._consumer
        self.reset()
        return consumer

    def digest_blueprint(self):
        """
        This method envokes blueprint digestion by the builder.
        It detects if a multistep build process is necessary and
        sets the channel build context.
        """
        LOGGER.debug('{BUILDER} digesting blueprint'.format(BUILDER=self.__class__.__name__))
        LOGGER.debug('{BUILDER} digesting blueprint OK!'.format(BUILDER=self.__class__.__name__))

    def start_independent_build(self):
        """
        This method invokes the standard building process for
        the consumer builder. The Foreman can invoke custom builds by
        instructing the builder on it's own.
        """
        self.setup_connection()
        self.setup_exchange()
        self.setup_queue()
        self.setup_callback_method()
        self.setup_qos()

    def setup_connection(self):
        """
        Set up the connection object for the consumer.
        """
        self._consumer.connection_ob = self.connection

    def setup_exchange(self):
        """
        Set up the exchange on the AMQP broker by declaring the exchange
        on the current context channel.
        """
        _exchange_name = self.blueprint['exchange']
        LOGGER.debug('Binding exchange {EXCHANGE_NAME} to producer'.format(
            EXCHANGE_NAME=_exchange_name
        )
        )
        self._consumer.exchange = _exchange_name
        LOGGER.debug('Binding exchange {EXCHANGE_NAME} to producer OK!'.format(
            EXCHANGE_NAME=_exchange_name
        )
        )

    def setup_queue(self):
        self._consumer.queue = self.blueprint['queue']

    def setup_callback_method(self):
        self._consumer.callback_method = self.blueprint['callback_method']

    def setup_qos(self):
        self._consumer.qos = self.blueprint['qos']


class AsynchronousConsumer:
    def __init__(self):
        self.connection_ob = None
        self.connection = None
        self.channel = None
        self.exchange = None
        self.queue = None
        self.callback_method = lambda body, properties: None
        self.qos = 1
        self.was_consuming = False
        self._consuming = False
        self._consumer_tag = None
        self._closing = False

    def connect(self):
        try:
            self.connection_ob.connect(
                on_open_callback=self.on_connection_open,
                on_open_error=self.on_connection_open_error,
                on_close_callback=self.on_connection_closed
            )
            self.connection = self.connection_ob.connection
        except Exception as e:
            LOGGER.exception("Trying to set connection attempt failed")
            pass

    def close_connection(self):
        self._consuming = False
        if self.connection.is_closing or self.connection.is_closed:
            LOGGER.debug("Connection is closing or already closed")
        else:
            LOGGER.debug("Closing connection")
            self.connection.close()

    def on_connection_open(self, _unused_connection):
        """
        This method is called by pika once the connection to the AMQP broker has
        been established. It passes the handle to the connection object in
        _unused_connection, now it unused but in this method mutations on
        the handle are possible.
        The method then calls the self.open_channel() method.
        :param _unused_connection: The connection.
        :type _unused_connection: pika.SelectConnection
        """
        LOGGER.debug("Connection opened")
        self.open_channel()

    def on_connection_open_error(self, _unused_connection, err):
        """
        This method is called by pika if the connection to the AMQP broker
        can"t be established. The self.reconnect() method is called.
        :param _unused_connection: The connection handle where error occurred.
        :type _unused_connection: pika.SelectConnection
        :param err: The error.
        :type err: Exception
        """
        LOGGER.error("Connection open failed, reconnecting: {ERROR}".format(ERROR=err))
        self.reconnect()

    def on_connection_closed(self, _unused_connection, reason):
        """
        This method is invoked by pika when the connection to the AMQP broker is
        closed unexpectedly. Since it is unexpected, this method will
        reconnect to the AMQP broker if it disconnects.
        :param _unused_connection: The closed connection handle.
        :type _unused_connection: pika.SelectConnection|Any
        :param reason: The exception representing the reason for loss of
            connection.
        :type reason: Exception
        """
        self.channel = None
        if self._closing:
            self.connection.ioloop.stop()
        else:
            LOGGER.warning("Connection closed, reconnect necessary: {REASON}".format(REASON=reason))
            self.reconnect()

    def reconnect(self):
        """
        This method will be invoked if the connection can"t be opened or is
        closed. Indicates that a reconnect is necessary then stops the
        ioloop.
        """
        self.should_reconnect = True
        self.stop()

    def open_channel(self):
        """
        This method will open a new channel with the AMQP broker by issuing the
        Channel. Open RPC command. When the AMQP broker confirms the channel is open
        by sending the Channel.OpenOK RPC reply, the on_channel_open method
        will be invoked.
        """
        LOGGER.debug("Creating a new channel")
        self.connection.channel(on_open_callback=self.on_channel_open)

    def on_channel_open(self, channel):
        """
        This method is invoked by pika when the channel has been opened.
        The channel object is passed in so this method can make use of it.
        Since the channel is now open, the exchange to use is declared.
        :param channel: The channel object.
        :type channel: pika.channel.Channel
        """
        LOGGER.debug("Channel opened")
        self.channel = channel
        self.add_on_channel_close_callback()
        self.set_qos()

    def add_on_channel_close_callback(self):
        """
        This method tells pika to call the on_channel_closed method if
        the AMQP broker unexpectedly closes the channel.
        """
        LOGGER.debug("Adding channel close callback")
        self.channel.add_on_close_callback(self.on_channel_closed)

    def on_channel_closed(self, channel, reason):
        """
        Invoked by pika when the AMQP broker unexpectedly closes the channel.
        Channels are usually closed if there is an attempt to do something that
        violates the protocol, such as re-declare an exchange or queue with
        different parameters. In this case, this method will close the
        connection to shutdown the object.
        :param channel: The closed channel.
        :type channel: pika.channel.Channel
        :param reason: The reason why the channel was closed.
        :type reason: Exception
        """
        LOGGER.warning("Channel {CHANNEL} was closed: {REASON}".format(CHANNEL=channel,
                                                                       REASON=reason))
        self.close_connection()

    def set_qos(self):
        """
        This method sets up the consumer prefetch to only be delivered
        one message at a time. The consumer must acknowledge this message
        before the AMQP broker will deliver another one. Do experiment
        with different prefetch values to achieve desired performance.
        """
        self.channel.basic_qos(
            prefetch_count=self.qos, callback=self.on_basic_qos_ok)

    def on_basic_qos_ok(self, _unused_frame):
        """
        This method is invoked by pika when the Basic.QoS method has completed.
        At this point the consumer will start consuming messages by calling the
        self.start_consuming() method which will invoke the needed RPC commands
        to start the process.
        :param _unused_frame: The Basic.QosOk response frame
        :type _unused_frame: pika.frame.Method
        """
        LOGGER.debug("QOS set to: {PREFETCH_COUNT}".format(PREFETCH_COUNT=self.qos))
        self.start_consuming()

    def start_consuming(self):
        """
        This method sets up the consumer by first calling
        add_on_cancel_callback so that the object is notified if the AMQP broker
        cancels the consumer. It then issues the Basic.Consume RPC command
        which returns the consumer tag that is used to uniquely identify the
        consumer with the AMQP broker. The method keeps the value globally to use
        it when we want to cancel consuming. The on_message method is passed
        in as a callback pika will invoke when a message is fully received.
        """
        LOGGER.debug("Issuing consumer related RPC commands")
        self.add_on_cancel_callback()
        self._consumer_tag = self.channel.basic_consume(
            queue=self.queue,
            on_message_callback=self.on_message
        )
        self.was_consuming = True
        self._consuming = True

    def on_message(self, channel, method, properties, body):
        """
        Invoked by pika when a message is delivered from the AMQP broker. The
        channel is passed for convenience. The basic_deliver object that
        is passed in carries the exchange, routing key, delivery tag and
        a redelivered flag for the message. The properties passed in is an
        instance of BasicProperties with the message properties and the body
        is the message that was sent.

        :param channel: The channel object.
        :type channel: pika.channel.Channel

        :param method: basic_deliver method.
        :type method: pika.Spec.Basic.Deliver

        :param properties: The properties.
        :type properties: pika.Spec.BasicProperties

        :param body: The message body.
        :type body: bytes
        """
        try:
            self.callback_method(json.loads(body), properties)
            self.acknowledge_message(method.delivery_tag)
        except Exception:
            LOGGER.exception("Asynchronous callback method exception:")

    # TODO: Integrate this as a choice
    def acknowledge_message(self, delivery_tag):
        self.channel.basic_ack(delivery_tag=delivery_tag)

    def add_on_cancel_callback(self):
        """
        Add a callback that will be invoked if the AMQP broker cancels the consumer
        for some reason. If the AMQP broker does cancel the consumer,
        on_consumer_cancelled will be invoked by pika.
        """
        LOGGER.debug("Adding consumer cancellation callback")
        self.channel.add_on_cancel_callback(self.on_consumer_cancelled)

    def on_consumer_cancelled(self, method_frame):
        """
        This method is invoked by pika when the AMQP broker sends a Basic.Cancel
        for a consumer receiving messages.
        :param method_frame: The Basic.Cancel frame
        :type method_frame: pika.frame.Method
        """
        LOGGER.debug("Consumer was cancelled remotely, shutting down: {METHOD_FRAME}".format(METHOD_FRAME=method_frame))
        if self.channel:
            self.channel.close()

    def stop_consuming(self):
        """
        Tell the AMQP broker that you would like to stop consuming by sending the
        Basic.Cancel RPC command.
        """
        if self.channel:
            LOGGER.debug("Sending a Basic.Cancel RPC command to the AMQP broker")
            cb = functools.partial(
                self.on_cancelok, userdata=self._consumer_tag)
            self.channel.basic_cancel(self._consumer_tag, cb)

    def on_cancelok(self, _unused_frame, userdata):
        """
        This method is invoked by pika when the AMQP broker acknowledges the
        cancellation of a consumer. At this point we will close the channel.
        This will invoke the on_channel_closed method once the channel has been
        closed, which will in-turn close the connection.
        :param _unused_frame: The Basic.CancelOk frame
        :type _unused_frame: pika.frame.Method
        :param userdata: Extra user data (consumer tag)
        :type userdata: str|unicode
        """
        self._consuming = False
        LOGGER.debug(
            "the AMQP broker acknowledged the cancellation of the consumer: {USERDATA}".format(USERDATA=userdata))
        self.close_channel()

    def close_channel(self):
        """
        Call to close the channel with the AMQP broker cleanly by issuing the
        Channel.Close RPC command.
        """
        LOGGER.debug("Closing the channel")
        self.channel.close()

    def start(self):
        """
        Start the consumer by connecting to the AMQP broker and then starting
        the IOLoop to block and allow the SelectConnection to operate.
        """
        self.connect()
        self.connection.ioloop.start()

    def stop(self):
        """
        Cleanly shutdown the connection to the AMQP broker by stopping the consumer
        with the AMQP broker. When the AMQP broker confirms the cancellation, on_cancelok
        will be invoked by pika, which will then closing the channel and
        connection. The IOLoop is started again because this method is invoked
        when raising a PoisonPill exception. This
        exception stops the IOLoop which needs to be running for pika to
        communicate with the AMQP broker. All of the commands issued prior to starting
        the IOLoop will be buffered but not processed.
        """
        if not self._closing:
            self._closing = True
            LOGGER.debug("Stopping")
            if self._consuming:
                self.stop_consuming()
                self.connection.ioloop.start()
            else:
                self.connection.ioloop.stop()
            LOGGER.debug("Stopped")


class SynchronousConsumerBuilder(AbstractSynchronousConsumerBuilder):

    def __init__(self, connection, blueprint):
        """
        This method initializes the consumer builder and sets the
        connection and the by the Foreman already on compatibility
        checked build template.

        :param connection: The provided AMQP broker connection that is attached
        to the Pokeman.
        :type connection: pokeman.Pokeman.connection

        :param blueprint: The resolved template from the Wingman.
        :type blueprint: pokeman.coating._resolvers.Wingman._blueprint
        """
        self.reset()
        self.connection = connection
        self.blueprint = blueprint

    def reset(self):
        """
        The production builder is reset on initialization and after a
        delivery of a consumer.
        """
        self._consumer = SynchronousConsumer()

    @property
    def consumer(self):
        """
        When the consumer delivery is invoked, the builder resets itself for
        a new producing task.

        :return: The built consumer object.
        :rtype: pokeman.coating._builders.SynchronousConsumer
        """
        consumer = self._consumer
        self.reset()
        return consumer

    def digest_blueprint(self):
        """
        This method envokes blueprint digestion by the builder.
        It detects if a multistep build process is necessary and
        sets the channel build context.
        """
        LOGGER.debug('{BUILDER} digesting blueprint'.format(BUILDER=self.__class__.__name__))
        LOGGER.debug('{BUILDER} digesting blueprint OK!'.format(BUILDER=self.__class__.__name__))

    def start_independent_build(self):
        """
        This method invokes the standard building process for
        the consumer builder. The Foreman can invoke custom builds by
        instructing the builder on it's own.
        """
        self.open_channel()
        self.setup_exchange()
        self.setup_queue()
        self.setup_callback_method()
        self.setup_qos()
        self.elevate_setup()
        self.setup_variation()

    def open_channel(self):
        """
        This method will open a new channel with AMQP broker and attach
        it to the producer.
        """
        LOGGER.debug('Opening a new channel')
        new_channel = self.connection.channel()
        self._consumer.channel = new_channel
        LOGGER.debug('Opening a new channel OK!')

    def setup_exchange(self):
        """
        Set up the exchange on the AMQP broker by declaring the exchange
        on the current context channel.
        """
        _exchange_name = self.blueprint['exchange']
        LOGGER.debug('Binding exchange {EXCHANGE_NAME} to producer'.format(
            EXCHANGE_NAME=_exchange_name
        )
        )
        self._consumer.exchange = _exchange_name
        LOGGER.debug('Binding exchange {EXCHANGE_NAME} to producer OK!'.format(
            EXCHANGE_NAME=_exchange_name
        )
        )

    def setup_queue(self):
        self._consumer.queue = self.blueprint['queue']

    def setup_callback_method(self):
        self._consumer.callback_method = self.blueprint['callback_method']

    def setup_qos(self):
        self._consumer.qos = self.blueprint['qos']

    def setup_variation(self):
        if 'reference' in self.blueprint:
            if 'correlation_id' in self.blueprint['reference']:
                print("SVRC")
                modified_consumer = self._consumer
                modified_consumer.correlation_id_reference = self.blueprint['reference']['correlation_id']['value']
                modified_consumer.start = self.blueprint['reference']['correlation_id']['start_method'].__get__(modified_consumer, type(modified_consumer))
                modified_consumer.on_message = self.blueprint['reference']['correlation_id']['on_message_method'].__get__(modified_consumer, type(modified_consumer))
                self._consumer = modified_consumer
            else:
                pass
        else:
            pass

    def elevate_setup(self):
        self._consumer.set_qos()


class SynchronousConsumer:
    def __init__(self):
        self.channel = None
        self.exchange = None
        self.queue = None
        self.callback_method = lambda body, properties: None
        self.qos = 1

    def set_qos(self):
        """
        This method sets up the consumer prefetch to only be delivered
        one message at a time. The consumer must acknowledge this message
        before the AMQP broker will deliver another one. Do experiment
        with different prefetch values to achieve desired performance.
        """
        self.channel.basic_qos(prefetch_count=self.qos)

    def start(self, count=-1):
        current_count = 0
        while current_count != count:
            current_count += 1
            self.start_consuming()

    def start_consuming(self):
        """
        This method issues the Basic.Consume RPC command
        which returns the consumer tag that is used to uniquely identify the
        consumer with the AMQP broker. The method keeps the value globally to use
        it when we want to cancel consuming. The on_message method is passed
        in as a callback pika will invoke when a message is fully received.
        """
        self.channel.basic_consume(
            queue=self.queue,
            on_message_callback=self.on_message
        )
        self.channel.start_consuming()

    def on_message(self, channel, method, properties, body):
        """
        Invoked by pika when a message is delivered from the AMQP broker. The
        channel is passed for convenience. The basic_deliver object that
        is passed in carries the exchange, routing key, delivery tag and
        a redelivered flag for the message. The properties passed in is an
        instance of BasicProperties with the message properties and the body
        is the message that was sent.

        :param channel: The channel object.
        :type channel: pika.channel.Channel

        :param method: basic_deliver method.
        :type method: pika.Spec.Basic.Deliver

        :param properties: The properties.
        :type properties: pika.Spec.BasicProperties

        :param body: The message body.
        :type body: bytes
        """
        try:
            self.callback_method(json.loads(body), properties)
            self.acknowledge_message(method.delivery_tag)
        except Exception:
            LOGGER.exception("Synchronous callback method exception:")
        self.channel.stop_consuming()

    # TODO: Integrate this as a choice
    def acknowledge_message(self, delivery_tag):
        self.channel.basic_ack(delivery_tag=delivery_tag)


# TODO: Add Template design pattern integration for Asynchonous Producer integration
class Foreman:
    """
    The Foreman gets it's instructions from the Pokeman and uses the
    Wingman for blueprint-builder validation, it can return builders
    after picking and assigning the right builders for the right blueprint
    and Ptypes.
    """
    def __init__(self):
        """
        This method initializes the Foreman with no attached builder and
        sets the Wingman as its validation servant.

        .. note::
            Pick and assign builders in chronological order, as the Foreman
            can only do the job allocation with one builder at a time.

            You can off-course deploy and operate multiple builders at the
            same time, as they handle the heavy lifting.
        """
        self.builder = None
        self.wingman = Wingman()

    def pick_builder(self, connection, coating, ptype):
        """
        This method picks the right builder for the provided coating
        and Ptypes. The Wingman resolves and validates the provided
        coating and supplies the blueprint.

        :param connection: The provided AMQP broker connection.
        :type connection: pokeman.Pokeman.connection

        :param coating: The provided EIP coating
        :type coating: pokeman.coating.BasicMessage

        :param ptype: The provided Ptypes
        :type ptype: pokeman.coating.ptypes.Ptypes
        """
        if self.builder is not None:
            raise SyntaxError('The {MANAGER} has already picked a builder. Assign some work to the builder first'
                              'or destroy the current builder with {MANAGER}.destroy_builder()'.format(
                MANAGER=self.__class__.__name__
            )
            )
        if coating.__class__.__name__ not in _c__all__:
            raise AttributeError('The provided coating is not valid')
        blueprint = self.wingman.resolve(coating=coating)
        ptype_check = Ptypes._map(eip=blueprint["type"], ptype=ptype)
        if ptype_check["map"] == Ptypes.SYNC_PRODUCER:
            self.builder = SynchronousProducerBuilder(connection=connection, blueprint=blueprint)
        elif ptype_check["map"] == Ptypes.SYNC_CONSUMER:
            self.builder = SynchronousConsumerBuilder(connection=connection, blueprint=blueprint)
        elif ptype_check["map"] == Ptypes.ASYNC_CONSUMER:
            self.builder = AsynchronousConsumerBuilder(connection=connection, blueprint=blueprint)
        else:
            raise NotImplementedError('The provided configuration parameters are not implemented yet')

    def deliver_producer(self):
        """
        This method invokes the building of a producer by the builder
        and delivers it to the Pokeman.

        :return: The producer object
        """
        self.builder.digest_blueprint()
        self.builder.start_independent_build()
        return self.builder.producer

    def deliver_consumer(self):
        """
        This method invokes the building of a consumer by the builder
        and delivers it to the Pokeman.

        :return: The consumer object
        """
        self.builder.digest_blueprint()
        self.builder.start_independent_build()
        return self.builder.consumer

    def destroy_builder(self):
        """
        This method destroys the current builder attached to the Foreman.
        Lunch-break for that guy.
        """
        self.builder = None



from pokeman.utils.custom_abc import ABCMeta, abstract_attribute


class AbstractBasicMessageConstruction(metaclass=ABCMeta):
    """
    Abstract base class for Enterprise Integration Patterns,
    in specific Message Constructions.
    """

    @abstract_attribute
    def app_id(self):
        pass

    @abstract_attribute
    def exchange(self):
        pass

    @abstract_attribute
    def routing_key(self):
        pass



from abc import ABC, abstractmethod
import inspect
from uuid import uuid4
import atexit
import os

from pokeman.amqp_resources import heapq as _heapq_
from pokeman.composite.connection import Connection, SelectConnection
from pokeman.coatings.ptypes import Ptypes
from pokeman.coatings.builders import Foreman
from pokeman.amqp_resources.builders import ResourceManager

import logging

LOGGER = logging.getLogger(__name__)


class AbstractPokeman(ABC):
    """
    Abstract Pokeman structure.
    """
    def __init__(self):
        """
        This method initializes the Pokeman setting the basic
        configuration parameters, creating the attached HeapQ
        and registering the Atexit method.
        """
        from pokeman import _current_os
        LOGGER.debug('Initializing Pokeman on current os: {OS}'.format(OS=_current_os))
        self.POKER_ID = str(uuid4())
        self.connection_parameters = None
        self.MSCID = 'main_{POKER_ID}'.format(POKER_ID=self.POKER_ID)
        self.connections = {
            'sync': {
                self.MSCID: None
            },
            'async': {}
        }

        self.MSC = lambda: self.connections['sync'][self.MSCID]
        self._declared = False
        self.channels = []
        self.cleaned_up = False
        _heapq_.ResourceHeapQ.create_database(poker_id=self.POKER_ID)
        atexit.register(self.cleanup)
        LOGGER.debug('Initializing Pokeman on current os: {OS} OK!'.format(OS=_current_os))

    @abstractmethod
    def set_parameters(self, connection):
        pass

    @abstractmethod
    def start(self):
        pass

    @abstractmethod
    def stop(self):
        pass

    @abstractmethod
    def apply_resources(self):
        pass

    @abstractmethod
    def declare_producer(self, coating, type):
        pass

    @abstractmethod
    def declare_consumer(self, coating, type):
        pass

    @abstractmethod
    def declare_router(self):
        pass

    @abstractmethod
    def declare_transformator(self):
        pass

    @abstractmethod
    def declare_channel_manager(self):
        pass

    @abstractmethod
    def declare_system_management_tool(self):
        pass

    @abstractmethod
    def cleanup(self):
        pass


class Pokeman(AbstractPokeman):
    """
    The main Pokeman class that handles all the composite resources by
    distributing tasks to the Foreman and indirect the Wingman.
    """
    def set_parameters(self, connection):
        self.connection_parameters = connection

    def set_sync_connection(self, main=False):
        if main:
            cid = self.MSCID
        else:
            cid = str(uuid4())
        self.connections['sync'][cid] = Connection(parameters=self.connection_parameters)
        return cid

    def set_async_connection(self):
        cid = str(uuid4())
        self.connections['async'][cid] = SelectConnection(parameters=self.connection_parameters)
        return cid

    def start(self):
        """
        This method starts the Pokeman by attaching the connection
        with the AMQP broker to itself.
        """
        if self.connection_parameters is not None:
            self.set_sync_connection(main=True)
            self.MSC().connect()
        else:
            raise AttributeError('No connection parameters set for the Pokeman.')

    def stop(self):
        """
        Stop the Pokeman by closing the channels that are attached to the Pokeman
        and the connection.
        """
        LOGGER.debug("Stopping Pokeman")
        for name, connection in self.connections['sync'].items():
            if connection is not None:
                connection.close_all_channels()
                connection.disconnect()
        for name, connection in self.connections['async'].items():
            if connection is not None:
                connection.close_all_channels()
                connection.disconnect()
        self.cleanup()
        LOGGER.debug("Stopping Pokeman OK!")

    def apply_resources(self):
        if self.MSC() is not None:
            _heapq_.ResourceHeapQ.apply_resources(connection=self.MSC().connection, poker_id=self.POKER_ID)
        else:
            raise ConnectionError('No active connection set. Make sure to start the Pokeman first.')
        
    def _set_coating_ptype_connection(self, ptype):
        """
        This method assigns the connection type depending on the
        synchronous or asynchronous nature of the coating Ptype.

        :param ptype: The provided Ptype
        :type ptype: Ptypes

        :return: The connection object

        .. note::
            Note that:
            * in case of a SYNC_ Ptype a live connection object is passed
            * in case of a ASYNC_ Ptype a not connected connection object
            is passed, because the asynchronous IOloop needs a callback
            waterfall.
            * in case of a ASYNC_ Ptype a new asynchronous connection
            object is created for each resource.
        """
        if self.MSC() is not None:
            if ptype.name.startswith('SYNC_'):
                return self.MSC().connection
            elif ptype.name.startswith('ASYNC_'):
                cid = self.set_async_connection()
                return self.connections['async'][cid]
        else:
            raise ConnectionError('No active connection set. Make sure to start the Pokeman first.')

    def declare_producer(self, coating, ptype):
        """
        This method invokes the building and delivery of a producer
        with the provided configuration managed by the Foreman.

        :param coating: The provided EIP coating
        :type coating: pokeman.coating

        :param ptype: The provided Ptypes
        :type ptype: pokeman.coating.ptypes.Ptypes

        :return: The producer
        """
        _coating = coating
        _coating.exchange = coating.exchange(_pkid=self)
        self.apply_resources()
        foreman = Foreman()
        foreman.pick_builder(
            connection=self._set_coating_ptype_connection(ptype=ptype),
            coating=_coating,
            ptype=ptype
        )
        producer = foreman.deliver_producer()
        if 'SYNC_' in ptype.name:
            self.channels.append(producer.channel)
        return producer

    def declare_consumer(self, coating, ptype):
        """
        This method invokes the building and delivery of a consumer
        with the provided configuration managed by the Foreman.

        :param coating: The provided EIP coating
        :type coating: pokeman.coating

        :param ptype: The provided Ptypes
        :type ptype: pokeman.coating.ptypes.Ptypes

        :return: The producer
        """
        _coating = coating
        _coating.exchange = coating.exchange(_pkid=self)
        self.apply_resources()
        foreman = Foreman()
        foreman.pick_builder(
            connection=self._set_coating_ptype_connection(ptype=ptype),
            coating=_coating,
            ptype=ptype
        )
        consumer = foreman.deliver_consumer()
        if 'SYNC_' in ptype.name:
            self.channels.append(consumer.channel)
        return consumer

    def declare_router(self):
        raise NotImplementedError(
            '{METHOD} method is not implemented yet for {CLASS} object'.format(
                METHOD=inspect.currentframe().f_code.co_name,
                CLASS=self.__class__.__name__
            )
        )

    def declare_transformator(self):
        raise NotImplementedError(
            '{METHOD} method is not implemented yet for {CLASS} object'.format(
                METHOD=inspect.currentframe().f_code.co_name,
                CLASS=self.__class__.__name__
            )
        )

    def declare_channel_manager(self):
        raise NotImplementedError(
            '{METHOD} method is not implemented yet for {CLASS} object'.format(
                METHOD=inspect.currentframe().f_code.co_name,
                CLASS=self.__class__.__name__
            )
        )

    def declare_system_management_tool(self):
        raise NotImplementedError(
            '{METHOD} method is not implemented yet for {CLASS} object'.format(
                METHOD=inspect.currentframe().f_code.co_name,
                CLASS=self.__class__.__name__
            )
        )

    def delete_attached_resources(self):
        """
        This method deletes all the AMQP resources attached to the Pokeman.

        ..note::
            Warning! Use this method with caution, this will delete all
            the resources attached to the pokeman, included the default
            resources if called from the default Pokeman. This method
            will delete resources that are still in use as well, and
            also deletes resources if they are re-declared in other
            Pokeman's as well but still have the same name.
        """
        if self.MSC() is not None:
            resource_manager = ResourceManager(connection=self.MSC().connection)
            resource_manager.delete_attached_resources(poker_id=self.POKER_ID)
        else:
            raise ConnectionError('No active connection set. Make sure to start the Pokeman first.')

    def cleanup(self):
        """
        Atexit and regular called cleanup method for the Pokeman object. Its primary
        function is to delete the HeapQ resources.
        """
        if self.connections is not None:
            if self.cleaned_up is False:
                try:
                    LOGGER.debug('Cleaning up Pokeman {POKER_ID}'.format(POKER_ID=self.POKER_ID))
                    # _heapq_.ResourceHeapQ.remove_heapq(poker_id=self.POKER_ID)
                    LOGGER.debug('Cleaning up Pokeman {POKER_ID} OK!'.format(POKER_ID=self.POKER_ID))
                    self.cleaned_up = True
                except FileNotFoundError:
                    pass
            else:
                pass
        else:
            raise ConnectionError('No active connection set. Make sure to start the Pokeman first.')


from inspect import getfullargspec

from pokeman.coatings.messaging_endpoints._abc_endpoint import AbstractBasicMessagingEndpoint
from pokeman.amqp_resources.globals import Exchange


class BasicMessageConsumer(AbstractBasicMessagingEndpoint):
    """
    Basic Message Enterprise Integration Pattern for messaging
    endpoint > Basic Message Consumer.

    A Sentinel is declared default for various pattern parameters,
    as the possible None declaration is a posteriori.
    """
    _SENTINEL = Ellipsis

    def __init__(
            self,
            exchange=_SENTINEL,
            queue=_SENTINEL,
            callback_method=_SENTINEL,
            qos=_SENTINEL
    ):
        """
        :param app_id: The provided app_id
        :type app_id: str

        :param exchange: The provided exchange.
        :type exchange: Exchange

        :param routing_key: The provided routing key.
        :type routing_key: pokeman.RoutingKey
        """
        self._exchange = exchange
        self._exchange_converter = lambda _pkid: Exchange(
            specific_poker=self._specific_poker_setter(
                specific_poker=_pkid,
                lambda_call=True
            )).exchange_name if self._exchange is self._SENTINEL else self._exchange.exchange_name
        self.exchange = self._exchange_setter(exchange=self._exchange_converter)
        self.queue = self._queue_setter(queue=queue)
        self.callback_method = self._callback_method_setter(callback_method=callback_method)
        self.qos = self._qos_setter(qos=qos)

    def _exchange_setter(self, exchange):
        return exchange

    def _exchange_name_setter(self, exchange_name):
        if exchange_name is self._SENTINEL:
            return "pokeman_exchange"
        else:
            return exchange_name

    def _queue_setter(self, queue):
        if queue is self._SENTINEL:
            return "pokeman_queue"
        else:
            return queue.queue_name

    def _callback_method_setter(self, callback_method):
        if callback_method is self._SENTINEL:
            return lambda body: 'No callback method provided'
        else:
            obligatory_args = ['body', 'properties']
            for arg in obligatory_args:
                if arg not in getfullargspec(callback_method).args:
                    raise AttributeError("The provided callback method needs to have the {ARG} argument".format(
                        ARG=arg
                    ))
            return callback_method

    def _qos_setter(self, qos):
        if qos is self._SENTINEL:
            return 1
        else:
            return qos

    def _specific_poker_setter(self, specific_poker, lambda_call=False):
        if specific_poker is self._SENTINEL:
            return None
        elif lambda_call is False:
            return specific_poker.poker_id
        elif lambda_call is True:
            return specific_poker
        else:
            raise ValueError('Unresolvable specific poker parameter')


import pytest
import unittest
from time import sleep

import logging

LOGGER = logging.getLogger(__name__)


@pytest.fixture(name="poker", scope="function")
def default_composite_poker():
    from pokeman import Pokeman, BasicConnection, BasicConfig
    from pokeman.amqp_resources.heapq import _POKEMAN_DEFAULT_HEAPQ_ENVIRON
    import os

    # Prevent pytest default HeapQ persistence
    if _POKEMAN_DEFAULT_HEAPQ_ENVIRON in os.environ:
        del os.environ[_POKEMAN_DEFAULT_HEAPQ_ENVIRON]

    composite_config = BasicConfig(
        connection_attempts=3,
        heartbeat=3600,
        retry_delay=1
    )
    composite = BasicConnection(
        connstr='amqp://guest:guest@localhost:5672',
        config=composite_config
    )
    poker = Pokeman()
    poker.connection_parameters(composite=composite)

    yield poker
    del poker


def test__pokeman_composite_connection(poker):
    from pika import BlockingConnection

    poker.start()
    assert isinstance(poker.connection, BlockingConnection)
    poker.stop()
    sleep(.05)


def test_basic_message_construction_coating(poker):
    from pokeman import BasicMessage
    from pokeman import Ptypes
    from pokeman import Exchange, RoutingKey, Queue

    # exchange = Exchange(exchange_name="my_exchange", exchange_type="direct")
    # routing_key = RoutingKey(key="my.key")
    # queue = Queue(queue_name="my_queue", exchange=exchange, routing_key=routing_key)
    rk = RoutingKey(key='my_messagekey')
    Queue(routing_key=rk)
    coating = BasicMessage(app_id="new-po", routing_key=rk)
    poker.start()
    producer_1 = poker.declare_producer(coating=coating, ptype=Ptypes.SYNC_PRODUCER)
    poker.apply_resources()
    # for _ in range(0, 1):
    #     producer_1.publish({"message": 123456})
    # for i in range(0, 10):
    #     producer_1.publish(message={"message": 123456})
    # sleep(6)
    poker.delete_attached_resources()
    poker.stop()

# def test_basic_messaging_endpoint_selective_consumer(poker):
#     from pokeman import BasicMessageConsumer
#     from pokeman import Ptypes
#     from pokeman import Exchange, Queue





from abc import ABC, abstractmethod

import json
from pika import BlockingConnection, SelectConnection
from pika.exceptions import ChannelClosedByBroker

from pokeman.amqp_resources.globals import AbstractExchange, AbstractQueue, Exchange, Queue
from . import __all__ as _c__all__

import logging

LOGGER = logging.getLogger(__name__)


class AbstractResourceBuilder(ABC):
    """
    Abstract base class for a resource builder.
    """
    @abstractmethod
    def digest_blueprint(self):
        pass

    @abstractmethod
    def create_resource(self):
        pass


class AbstractResourceDestroyer(ABC):
    """
    Abstract base class for a resource destroyer.
    """
    def __init__(self, channel, blueprint):
        self.channel = channel
        self.blueprint = blueprint
        self.blueprint_context = self.blueprint['context']
        self.delete_resource()

    @abstractmethod
    def delete_resource(self):
        pass


class AbstractExistenceChecker(ABC):
    """
    Abstract base class for an existence checker.
    """
    def __init__(self, channel, blueprint):
        self.reset()
        self.channel = channel
        self.blueprint = blueprint
        self.blueprint_context = self.blueprint['context']
        self.check_existence()

    def reset(self):
        self._exists = False

    @property
    def exists(self):
        existence = self._exists
        self.reset()
        return existence

    @abstractmethod
    def check_existence(self):
        pass


class ExchangeBuilder(AbstractResourceBuilder, AbstractExchange):
    """
    The Exchange builder is responsible for building Queue resources
    and is handled by the ResourceManager.
    """
    def __init__(self, channel, blueprint):
        """
        This method initializes the Exchange builder.

        :param channel: The provided AMQP channel.
        :type channel: pokeman.Pokeman.connection.channel

        :param blueprint: The provided Exchange blueprint
        :type blueprint: dict
        """
        self.reset()
        self.channel = channel
        self.blueprint = blueprint
        self.build_context = self.blueprint['context']

        # Abstract attributes
        self.exchange_name = None
        self.exchange_type = None
        self.durable = None
        self.auto_delete = None
        self.specific_poker = None

    def reset(self):
        """
        This method resets the Exchange builder instance.
        """
        self.exchange_name = None
        self.exchange_type = None
        self.durable = None
        self.auto_delete = None
        self.specific_poker = None

    def digest_blueprint(self):
        """
        This method checks whether the right blueprint was supplied to
        the builder and copies the build context parameters to its
        instance.
        """
        LOGGER.debug('{BUILDER} digesting blueprint'.format(BUILDER=self.__class__.__name__))
        if self.blueprint['type'] != self.__class__.__name__.replace('Builder', ''):
            raise ValueError('{BUILDER} received the wrong blueprint'.format(
                BUILDER=self.__class__.__name__
            ))
        self.exchange_name = self.build_context['exchange_name']
        self.exchange_type = self.build_context['exchange_type']
        self.durable = self.build_context['durable']
        self.auto_delete = self.build_context['auto_delete']
        self.specific_poker = self.build_context['specific_poker']
        LOGGER.debug('{BUILDER} digesting blueprint OK!'.format(BUILDER=self.__class__.__name__))

    def create_resource(self):
        """
        This method creates the Exchange resource on the AMQP broker.
        """
        LOGGER.debug('Creating new Exchange')
        self.channel.exchange_declare(
            exchange=self.exchange_name,
            exchange_type=self.exchange_type,
            durable=self.durable,
            auto_delete=False
        )
        LOGGER.debug('Creating new Exchange OK!')
        self.reset()


class QueueBuilder(AbstractResourceBuilder, AbstractQueue):
    """
    The Queue builder is responsible for building Queue resources
    and is handled by the ResourceManager.
    """
    def __init__(self, channel, blueprint):
        """
        This method initializes the Queue builder.

        :param channel: The provided AMQP channel.
        :type channel: pokeman.Pokeman.connection.channel

        :param blueprint: The provided Queue blueprint
        :type blueprint: dict
        """
        self.reset()
        self.channel = channel
        self.blueprint = blueprint
        self.build_context = self.blueprint['context']

        # Abstract attributes
        self.queue_name = None
        self.exchange = None
        self.routing_key = None
        self.exclusive = None
        self.durable = None
        self.auto_delete = None
        self.specific_poker = None

    def reset(self):
        """
        This method resets the Queue builder instance.
        """
        self.queue_name = None
        self.exchange = None
        self.routing_key = None
        self.exclusive = None
        self.durable = None
        self.auto_delete = None
        self.specific_poker = None

    def digest_blueprint(self):
        """
        This method checks whether the right blueprint was supplied to
        the builder and copies the build context parameters to its
        instance.
        """
        LOGGER.debug('{BUILDER} digesting blueprint'.format(BUILDER=self.__class__.__name__))
        if self.blueprint['type'] != self.__class__.__name__.replace('Builder', ''):
            raise ValueError('{BUILDER} received the wrong blueprint'.format(
                BUILDER=self.__class__.__name__
            ))
        self.queue_name = self.build_context['queue_name']
        self.exchange = self.build_context['exchange']
        self.routing_key = self.build_context['routing_key']
        self.exclusive = self.build_context['exclusive']
        self.durable = self.build_context['durable']
        self.auto_delete = self.build_context['auto_delete']
        LOGGER.debug('{BUILDER} digesting blueprint OK!'.format(BUILDER=self.__class__.__name__))

    def create_resource(self):
        """
        This method creates the Queue resource on the AMQP broker.
        """
        LOGGER.debug('Creating new Queue')
        self.channel.queue_declare(
            queue=self.queue_name,
            durable=self.durable,
            exclusive=self.exclusive,
            auto_delete=self.auto_delete
        )
        self.channel.queue_bind(
            queue=self.queue_name,
            exchange=self.exchange,
            routing_key=self.routing_key
        )
        LOGGER.debug('Creating new Queue')
        self.reset()


class ExchangeExistenceChecker(AbstractExistenceChecker):
    """
    The Exchange existence checker.
    """
    def check_existence(self):
        """
        This method checks if the provided resource already exists.

        :return: The existence status
        :rtype: bool
        """
        try:
            self.channel.exchange_declare(
                exchange=self.blueprint_context['exchange_name'],
                exchange_type=self.blueprint_context['exchange_type'],
                passive=True
            )
            self._exists = True
        except ChannelClosedByBroker as exc:
            if 'NOT_FOUND' in exc.reply_text:
                self._exists = False  # pragma: no cover


class QueueExistenceChecker(AbstractExistenceChecker):
    """
    The Queue existence checker.
    """
    def check_existence(self):
        """
        This method checks if the provided resource already exists.

        :return: The existence status
        :rtype: bool
        """
        try:
            self.channel.queue_declare(
                queue=self.blueprint_context['queue_name'],
                passive=True
            )
            self._exists = True
        except ChannelClosedByBroker as exc:
            if 'NOT_FOUND' in exc.reply_text:
                self._exists = False  # pragma: no cover


class ExchangeDestroyer(AbstractResourceDestroyer):
    """
    The Exchange destroyer object.
    """
    def delete_resource(self):
        """
        This method deletes the Exchange.
        """
        LOGGER.debug('Deleting Exchange')
        try:
            self.channel.exchange_delete(exchange=self.blueprint_context['exchange_name'])
        except ChannelClosedByBroker as exc:
            if 'NOT_FOUND' in exc.reply_text:
                pass
        LOGGER.debug('Deleting Exchange OK!')


class QueueDestroyer(AbstractResourceDestroyer):
    """
    The Queue destroyer object.
    """
    def delete_resource(self):
        """
        This methods deletes the Queue.
        """
        LOGGER.debug('Deleting Queue')
        try:
            self.channel.queue_delete(queue=self.blueprint_context['queue_name'])
        except ChannelClosedByBroker as exc:
            if 'NOT_FOUND' in exc.reply_text:
                pass
        LOGGER.debug('Deleting Queue OK!')


class ResourceManager:
    """
    The ResourceManager handles the validation, building and deletion
    of AMQP resources.
    """
    _templates = {
        Exchange.__name__: {
            'builder': ExchangeBuilder,
            'checker': ExchangeExistenceChecker,
            'destroyer': ExchangeDestroyer
        },
        Queue.__name__: {
            'builder': QueueBuilder,
            'checker': QueueExistenceChecker,
            'destroyer': QueueDestroyer
        }
    }

    def __init__(self, connection):
        """
        This method initialized the AMQP ResourceManager with no attached builder.

        As there is a hight risk of losing the connection during the various validation
        and building steps, each job is provided with its own channel derived from
        the Pokeman connection.

        :param connection: The provided connection
        :type connection: pokeman.Pokeman.connection
        """
        self.builder = None
        self._connection = connection
        self.resource_exists = False

    def get_channel(self):
        channel = None
        if self._connection.__class__.__name__ == BlockingConnection.__name__:
            channel = self._connection.channel()
        elif self._connection.__class__.__name__ == SelectConnection.__name__:
            self._connection.channel(on_open_callback=lambda ch: ch)
        return channel

    def pick_builder(self, blueprint):
        """
        This method picks the right builder for the provided resource blueprint.
        The ResourceManager gives the call to check if the resource already exists
        and assigns the right builder for the job.

        :param blueprint: The provided blueprint.
        :type blueprint: dict
        """
        if self.builder is not None:
            raise SyntaxError('The {MANAGER} has already picked a builder. Assign some work to the builder first'
                              'or destroy the current builder with Foreman.destroy_builder()'.
                              format(MANAGER=self.__class__.__name__)
                              )
        if blueprint['type'] not in _c__all__:
            raise ValueError('The provided resource is not valid')

        self.resource_exists = self._templates[blueprint['type']]['checker'](
            channel=self.get_channel(),
            blueprint=blueprint
        ).exists

        if self.resource_exists is False:
            self.builder = self._templates[blueprint['type']]['builder'](
                channel=self.get_channel(),
                blueprint=blueprint
            )
        else:
            LOGGER.debug('The provided resource already exists')

    def create_resource(self):
        """
        This method instructs the attached builder to create the resource.
        """
        if self.resource_exists is False:
            self.builder.digest_blueprint()
            self.builder.create_resource()
            self.destroy_builder()
        else:
            self.destroy_builder()
            pass

    def destroy_builder(self):
        """
        This method destroys the builder that is attached to the ResourceManager.
        """
        self.builder = None
        self.resource_exists = False

    def delete_attached_resources(self, poker_id):
        """
        This method deletes all the AMQP resources attached to the Pokeman.

        ..note::
            Warning! Use this method with caution, this will delete all
            the resources attached to the pokeman, included the default
            resources if called from the default Pokeman. This method
            will delete resources that are still in use as well, and
            also deletes resources if they are re-declared in other
            Pokeman's as well but still have the same name.

        :param poker_id: The provided poker id.
        :type poker_id: str
        """
        from pokeman.amqp_resources.heapq import ResourceHeapQ  # Avoid circular import
        LOGGER.debug('Deleting all attached resources for Pokeman {POKER_ID}'.format(
            POKER_ID=poker_id
        )
        )
        resources_heapq = ResourceHeapQ.get_heapq(
            poker_id=poker_id,
            reverse=True
        )
        for _ in range(0, resources_heapq.qsize()):
            heap_record = resources_heapq.get()
            resource_blueprint = json.loads(heap_record[1])
            LOGGER.debug('DELETING: ' + str(resource_blueprint))
            self._templates[resource_blueprint['type']]['destroyer'](
                channel=self._connection.channel(),
                blueprint=resource_blueprint
            )
        LOGGER.debug('Deleting all attached resources for Pokeman {POKER_ID} OK!'.format(
            POKER_ID=poker_id)
        )



import unittest

from pokeman.utils import task


class TaskStatusTests(unittest.TestCase):
    def test_base_name(self):
        self.assertEqual(task.TaskStatus.__name__, 'TaskStatus')

    def test_enums(self):
        self.assertEqual(task.TaskStatus.PENDING.name, 'PENDING')
        self.assertEqual(task.TaskStatus.SUCCESS.name, 'SUCCESS')
        self.assertEqual(task.TaskStatus.ON_HOLD.name, 'ON_HOLD')
        self.assertEqual(task.TaskStatus.FAILED.name, 'FAILED')

if __name__ == '__main__':
    unittest.main()



import unittest
import os
import sqlite3

from pokeman.amqp_resources import heapq


class ResourceHeapQTests(unittest.TestCase):
    POKER_ID_MOCK = '4ff44fd6-1d08-49a3-a98b-27073871e85c'
    HEAPQ_ID_MOCK = '{PREFIX}{POKER_ID}'.format(
        PREFIX=heapq._HEAPQ_PREFIX,
        POKER_ID=POKER_ID_MOCK
    )

    def setUp(self):
        self.ResourceHeapQ = heapq.ResourceHeapQ

    def tearDown(self):
        os.remove(self.ResourceHeapQ._set_heapq_location(heapq_id=self.HEAPQ_ID_MOCK))

    def test_base_name(self):
        self.assertEqual(self.ResourceHeapQ.__name__, 'ResourceHeapQ')

    def test__set_heapq_id(self):
        poker_id = self.POKER_ID_MOCK
        desired_heapq_id = '{PREFIX}{POKER_ID}'.format(
            PREFIX=heapq._HEAPQ_PREFIX,
            POKER_ID=poker_id
        )
        heapq_id = self.ResourceHeapQ._set_heapq_id(
            poker_id=poker_id
        )
        self.assertEqual(heapq_id, desired_heapq_id)

    def test__set_heapq_location(self):
        heapq_location = self.ResourceHeapQ._set_heapq_location(heapq_id=self.HEAPQ_ID_MOCK)
        directory_path = heapq_location.split('heapq_')[0]
        self.assertTrue(os.path.isdir(directory_path))

    def test__set_heapq_environment_variable_default_heapq(self):
        default_heapq = self.ResourceHeapQ._set_heapq_environment_variable(self.HEAPQ_ID_MOCK)
        self.assertEqual(os.environ[heapq._POKEMAN_DEFAULT_HEAPQ_ENVIRON], self.HEAPQ_ID_MOCK)
        self.assertTrue(default_heapq)

    def test__set_heapq_environment_variable_non_default_heapq(self):
        """Depends on previous test"""
        default_heapq = self.ResourceHeapQ._set_heapq_environment_variable(self.HEAPQ_ID_MOCK)
        self.assertFalse(default_heapq)

    def test_heapq_connect(self):
        heapq_location = self.ResourceHeapQ._set_heapq_location(heapq_id=self.HEAPQ_ID_MOCK)
        heapq_connection = self.ResourceHeapQ._heapq_connect(heapq_location=heapq_location)
        self.assertTrue(isinstance(heapq_connection, sqlite3.Connection))

    def test_create_database(self):
        self.assertEqual(self.ResourceHeapQ.create_database(poker_id=self.POKER_ID_MOCK), None)





# import pytest
#
# from pokeman import Queue, Exchange
#
# import os
# import time
#
#
# @pytest.fixture(name="poker", scope="function")
# def default_composite_poker():
#     from pokeman._pokeman import Pokeman
#     from pokeman import composite
#
#     composite_config = composite.BasicConfig(
#         connection_attempts=3,
#         heartbeat=3600,
#         retry_delay=1
#     )
#     enterprise_service_bus = composite.BlockingESB(
#         connstr='amqp://guest:guest@localhost:5672',
#         config=composite_config
#     )
#     poker = Pokeman()
#     poker.connection_parameters(composite=enterprise_service_bus)
#
#     yield poker
#
#
# def test_heapq(poker):
#
#     from pokeman._pokeman import Pokeman
#     from pokeman import composite
#
#     composite_config = composite.BasicConfig(
#         connection_attempts=3,
#         heartbeat=3600,
#         retry_delay=1
#     )
#     enterprise_service_bus = composite.BlockingESB(
#         connstr='amqp://guest:guest@localhost:5672',
#         config=composite_config
#     )
#     poker_a = Pokeman()
#     print(poker_a.poker_id)
#     poker_a.connection_parameters(composite=enterprise_service_bus)
#     poker_b = Pokeman()
#     print(poker_b.poker_id)
#     poker_b.connection_parameters(composite=enterprise_service_bus)
#
#     exchange_a = Exchange(exchange_name='aaedsadx')
#     exchange_b = Exchange(exchange_name='bbex')
#     Queue(queue_name='aadssaaa', durable=False, auto_delete=True, specific_poker=poker_a, exchange=exchange_a)
#     Queue(queue_name='bbsdsdabb', durable=False, auto_delete=True, specific_poker=poker_b, exchange=exchange_b)
#     exchange_c = Exchange(exchange_name='aatdasdron', durable=False)
#     poker_c = Pokeman()
#     exchange_d = Exchange(exchange_name='cctasdron', durable=False, specific_poker=poker_c)
#     poker_c.connection_parameters(composite=enterprise_service_bus)
#
#     poker_a.start()
#     poker_b.start()
#     poker_a.apply_resources()
#     poker_b.apply_resources()
#     poker_c.start()
#     poker_c.apply_resources()
#
#     poker_c.stop()
#     poker_a.stop()
#     poker_b.stop()
#



from .message_constructions.message import BasicMessage
from .messaging_endpoints.basic_message_consumer import BasicMessageConsumer
from .messaging_endpoints.polling_consumer import PollingConsumer
from .messaging_endpoints.selective_consumer import SelectiveConsumer
from .ptypes import Ptypes

__all__ = [
    # General
    'Ptypes',

    # Message Constructions
    'BasicMessage',

    # Messaging Endpoints
    'PollingConsumer', 'BasicMessageConsumer', 'SelectiveConsumer'

]



from enum import Enum


class Ptypes(Enum):

    ASYNC_PRODUCER = 0x10
    SYNC_PRODUCER = 0x20

    ASYNC_CONSUMER = 0x30
    SYNC_CONSUMER = 0x40

    __BasicMessage__ = ~ASYNC_PRODUCER & SYNC_PRODUCER
    __PollingConsumer__ = ASYNC_CONSUMER & ~SYNC_CONSUMER
    __BasicMessageConsumer__ = ~ASYNC_CONSUMER & SYNC_CONSUMER
    __SelectiveConsumer__ = ~ASYNC_CONSUMER & SYNC_CONSUMER

    @classmethod
    def _map(cls, eip, ptype):
        if '__{EIP}__'.format(EIP=eip) not in cls.__dict__.keys():
            raise KeyError('The provided coating has no mapping configuration')

        elif ptype is None or ptype.name not in cls.__dict__.keys():
            raise KeyError('The provided Ptype does not exist')

        elif cls.__dict__['__{EIP}__'.format(EIP=eip)] != ptype.value:
            raise ValueError('The provided coating is not compatible with the provided Ptype.\n'
                             'Check the mapping compatibility with Ptypes.option_check()')

        elif '__{EIP}__'.format(EIP=eip) in cls.__dict__.keys() and ptype.name in cls.__dict__.keys():
            return dict(map=ptype)

        else:
            raise ValueError('The provided coating and Ptype could not be resolved')

    @classmethod
    def option_check(cls, eip):
        if cls.__dict__['__{EIP}__'.format(EIP=eip)] == -0xf1:
            return dict(options=None)

        elif cls.__dict__['__{EIP}__'.format(EIP=eip)] == 0xa0:
            return dict(options=[cls.SYNC_PRODUCER])

        elif cls.__dict__['__{EIP}__'.format(EIP=eip)] == 0x50:
            return dict(options=[cls.ASYNC_PRODUCER])

        elif cls.__dict__['__{EIP}__'.format(EIP=eip)] == 0x0:
            return dict(options=[cls.SYNC_PRODUCER, cls.ASYNC_PRODUCER])



from abc import ABC, abstractmethod
from enum import Enum, IntFlag, auto

from pokeman.amqp_resources.globals import Exchange, Queue
from pokeman.amqp_resources import __all__ as _c__all__


class Action(Enum):
    """
    This Enum object hold the possible resource handling actions.
    """
    CREATE = 0x1
    DELETE = 0x0


class Priority(IntFlag):
    """
    Priority flags are set auto 8-bit integers.

    .. note::
        The assigned integer flags are processed
        bij the RESOURCE_HEAP_Q
        queue.
    """
    EXCHANGE = auto()
    QUEUE = auto()


class AbstractResourceResolver(ABC):
    """
    Abstract base clase for a AMQP resource resolver.
    """
    def __init__(self, resource, action):
        """
        :param resource: The provided resource.
        :type resource: pokeman._components.globals.*
        :param action: The action.
        :type action: Actions
        """
        self.resource = resource
        self.blueprint = {
            'type': None,
            'priority': None,
            'context': {},
            'action': action.name
        }
        self.resolve_resource()

    def resolve_resource(self):
        """
        This method executes all the resolve steps.
        """
        self.get_build_type()
        self.set_priority()
        self.check_configuration()
        self.get_context()

    def get_build_type(self):
        self.blueprint['type'] = self.resource.__class__.__name__

    @abstractmethod
    def set_priority(self):
        pass

    @abstractmethod
    def check_configuration(self):
        pass

    @abstractmethod
    def get_context(self):
        pass


class ExchangeResolver(AbstractResourceResolver):
    """
    Template resolver for Exchange.
    """
    def set_priority(self):
        """
        This method sets the priority value that is needed in the HeapQ,
        for the resource.
        """
        self.blueprint['priority'] = Priority.EXCHANGE.value

    def check_configuration(self):
        """
        This method does nothing yet.
        """
        pass

    def get_context(self):
        """
        This method lazy assigns all the resource attributes to
        the blueprint.

        ..note::
            Private '_' attributes are ignored.
        """
        for _attr, _value in self.resource.__dict__.items():
            if _attr.startswith('_'):
                continue
            self.blueprint['context'][_attr] = _value


class QueueResolver(AbstractResourceResolver):
    """
    Template resolver for Queue.
    """
    def set_priority(self):
        """
        This method sets the priority value that is needed in the HeapQ,
        for the resource.
        """
        self.blueprint['priority'] = Priority.QUEUE.value

    def check_configuration(self):
        """
        This method does nothing yet.
        """
        pass

    def get_context(self):
        """
        This method lazy assigns all the resource attributes to
        the blueprint.

        ..note::
            Private '_' attributes are ignored.
        """
        for _attr, _value in self.resource.__dict__.items():
            if _attr.startswith('_'):
                continue
            self.blueprint['context'][_attr] = _value


class ResourceResolver:
    """
    The resource resolver provides blueprints for the provided
    resource objects
    """
    _templates = {
        Exchange: {
            'blueprint': ExchangeResolver
        },
        Queue: {
            'blueprint': QueueResolver
        }
    }

    def __init__(self):
        """
        This method resets the ResourceResolver on initialization.
        """
        self.reset()

    def reset(self):
        """
        This method resets the blueprint object to None.
        """
        self._blueprint = None

    def resolve(self, resource, action):
        """
        This method takes the resource object and desired action,
        resolves it to a blueprint for builder objects.

        :param resource: The provided AMQP resource.
        :type resource: pokeman.amqp_resources.globals.*

        :param action: The provided action.
        :type action: Action

        :return: The resolved resource as blueprint.
        :rtype: dict

        ..note::
            This layer method calls the blueprint method from this class,
            as for future integration more checks and balances will be
            added.
        """
        blueprint = self.blueprint(resource=resource, action=action)
        return blueprint

    def blueprint(self, resource, action):
        """
        This method takes the resource object and desired action,
        resolves it to a blueprint for builder objects.

        :param resource: The provided AMQP resource.
        :type resource: pokeman.amqp_resources.globals.*

        :param action: The provided action.
        :type action: Action

        :return: The resolved resource as blueprint.
        :rtype: dict
        """
        if resource.__class__ in self._templates.keys():
            _resource = resource
        else:
            raise ValueError('No resource resolver implemented for {RESOURCE}'.format(
                RESOURCE=resource.__class__.__name__)
            )

        blueprint_resolver = self._templates[_resource.__class__]['blueprint'](resource=resource, action=action)
        self._blueprint = blueprint_resolver.blueprint
        blueprint = self._blueprint
        self.reset()
        return blueprint



from pokeman.coatings.messaging_endpoints.basic_message_consumer import BasicMessageConsumer


class SelectiveConsumer(BasicMessageConsumer):
    """
    Basic Message Enterprise Integration Pattern for messaging
    endpoint > Selective Consumer.

    A Sentinel is declared default for various pattern parameters,
    as the possible None declaration is a posteriori.
    """
    _SENTINEL = Ellipsis

    def __init__(
            self,
            exchange=_SENTINEL,
            queue=_SENTINEL,
            callback_method=_SENTINEL,
            qos=_SENTINEL,
            correlation_id=_SENTINEL
    ):
        super().__init__(
            exchange=exchange,
            queue=queue,
            callback_method=callback_method,
            qos=qos
        )
        self.correlation_id = self._correlation_id_setter(correlation_id=correlation_id)

    def _correlation_id_setter(self, correlation_id):
        if correlation_id is self._SENTINEL:
            raise ValueError('The SelectiveConsumer needs a selective reference: correlation_id')
        else:
            return correlation_id



import unittest

from tests import TestAttributes as TA

from pokeman.amqp_resources import globals


class AbstractExchangeTests(unittest.TestCase):
    def setUp(self):
        self.AbstractExchange = globals.AbstractExchange

    def test_base_name(self):
        self.assertEqual(self.AbstractExchange.__name__, 'AbstractExchange')

    def test_base_abstract_attributes(self):
        assert self.AbstractExchange.exchange_name
        self.AbstractExchange.exchange_name = None
        assert self.AbstractExchange.exchange_type
        self.AbstractExchange.exchange_type = None
        assert self.AbstractExchange.durable
        self.AbstractExchange.durable = None
        assert self.AbstractExchange.auto_delete
        self.AbstractExchange.auto_delete = None
        assert self.AbstractExchange.specific_poker
        self.AbstractExchange.specific_poker = None
        self.assertTrue(isinstance(self.AbstractExchange(), globals.AbstractExchange))


class AbstractRoutingKeyTests(unittest.TestCase):
    def setUp(self):
        self.AbstractRoutingKey = globals.AbstractRoutingKey

    def test_base_name(self):
        self.assertEqual(self.AbstractRoutingKey.__name__, 'AbstractRoutingKey')

    def test_base_abstract_attributes(self):
        assert self.AbstractRoutingKey.key
        self.AbstractRoutingKey.key = None
        self.assertTrue(isinstance(self.AbstractRoutingKey(), globals.AbstractRoutingKey))


class AbstractQueueTests(unittest.TestCase):
    def setUp(self):
        self.AbstractQueue = globals.AbstractQueue

    def test_base_name(self):
        self.assertEqual(self.AbstractQueue.__name__, 'AbstractQueue')

    def test_base_abstract_attributes(self):
        assert self.AbstractQueue.queue_name
        self.AbstractQueue.queue_name = None
        assert self.AbstractQueue.exchange
        self.AbstractQueue.exchange = None
        assert self.AbstractQueue.routing_key
        self.AbstractQueue.routing_key = None
        assert self.AbstractQueue.exclusive
        self.AbstractQueue.exclusive = None
        assert self.AbstractQueue.durable
        self.AbstractQueue.durable = None
        assert self.AbstractQueue.auto_delete
        self.AbstractQueue.auto_delete = None
        assert self.AbstractQueue.specific_poker
        self.AbstractQueue.specific_poker = None
        self.assertTrue(isinstance(self.AbstractQueue(), globals.AbstractQueue))


class ExchangeTests(unittest.TestCase):
    class PokemanMock:
        def __init__(self):
            self.poker_id = 'xyz'

    def setUp(self):
        from pokeman.amqp_resources.heapq import ResourceHeapQ
        with TA.patch(ResourceHeapQ, 'add_resource', lambda resource, specific_poker: None):
            self.exchange_set = globals.Exchange(
                exchange_name='test_name',
                exchange_type='test_type',
                durable=False,
                auto_delete=True,
                specific_poker=self.PokemanMock()
            )
            self.exchange_unset = globals.Exchange()

    def test_base_name(self):
        self.assertEqual(self.exchange_set.__class__.__name__, 'Exchange')

    def test__exchange_name_setter(self):
        self.assertEqual(self.exchange_set.exchange_name, 'test_name')
        self.assertEqual(self.exchange_unset.exchange_name, 'pokeman_exchange')

    def test__exchange_type_setter(self):
        self.assertEqual(self.exchange_set.exchange_type, 'test_type')
        self.assertEqual(self.exchange_unset.exchange_type, 'direct')

    def test__durable_setter(self):
        self.assertEqual(self.exchange_set.durable, False)
        self.assertEqual(self.exchange_unset.durable, True)

    def test__auto_delete_setter(self):
        self.assertEqual(self.exchange_set.auto_delete, True)
        self.assertEqual(self.exchange_unset.auto_delete, False)

    def test__specific_poker_setter(self):
        from pokeman import Queue
        self.assertEqual(self.exchange_set.specific_poker, 'xyz')
        self.assertEqual(self.exchange_unset.specific_poker, None)

class RoutingKeyTests(unittest.TestCase):
    def setUp(self):
        self.routingkey_set = globals.RoutingKey(key='xyz')
        self.routingkey_unset = globals.RoutingKey()

    def test_base_name(self):
        self.assertEqual(self.routingkey_set.__class__.__name__, 'RoutingKey')

    def test__routing_key_setter(self):
        self.assertEqual(self.routingkey_set.key, 'xyz')
        self.assertEqual(self.routingkey_unset.key, None)


class QueueTests(unittest.TestCase):
    class PokemanMock:
        def __init__(self):
            self.poker_id = 'xyz'

    class ExchangeMock:
        def __init__(self, state='set'):
            if state == 'set':
                self.exchange_name = 'mock_test_name'
            else:
                self.exchange_name = 'lol'

    class ExchangeConverterMock:
        def __init__(self, specific_poker):
            self.specific_poker = specific_poker
            self.exchange_name = 'converted_mock_test_name'

    class RoutingKeyMock:
        def __init__(self):
            self.key = 'mock_routing_key'

    class RoutingKeyConverterMock:
        def __init__(self):
            self.key = 'converted_mock_routing_key'

    def setUp(self):
        from pokeman.amqp_resources.heapq import ResourceHeapQ
        with TA.patch(ResourceHeapQ, 'add_resource', lambda resource, specific_poker: None):
            self.queue_set = globals.Queue(
                queue_name='test_name',
                exchange=self.ExchangeMock(state='set'),
                routing_key=self.RoutingKeyMock(),
                exclusive=True,
                durable=False,
                auto_delete=True,
                specific_poker=self.PokemanMock()
            )
            self.queue_unset = globals.Queue()

    def test_base_name(self):
        self.assertEqual(self.queue_set.__class__.__name__, 'Queue')

    def test__exchange_converter_lambda(self):
        with TA.patch(globals, 'Exchange', self.ExchangeConverterMock):
            self.assertEqual(self.queue_set._exchange_converter(), 'mock_test_name')
            self.assertEqual(self.queue_unset._exchange_converter(), 'converted_mock_test_name')

    def test__routing_key_converter_lambda(self):
        with TA.patch(globals, 'RoutingKey', self.RoutingKeyConverterMock):
            self.assertEqual(self.queue_set._routing_key_converter(), 'mock_routing_key')
            self.assertEqual(self.queue_unset._routing_key_converter(), 'converted_mock_routing_key')

    def test__queue_name_setter(self):
        self.assertEqual(self.queue_set.queue_name, 'test_name')
        self.assertEqual(self.queue_unset.queue_name, 'pokeman_queue')

    def test__exchange_setter(self):
        """
        Checked with above lambda test.
        """
        pass

    def test__routing_key_setter(self):
        """
        Checked with above lambda test.
        """
        pass

    def test__exclusive_setter(self):
        self.assertEqual(self.queue_set.exclusive, True)
        self.assertEqual(self.queue_unset.exclusive, False)

    def test__durable_setter(self):
        self.assertEqual(self.queue_set.durable, False)
        self.assertEqual(self.queue_unset.durable, True)

    def test__auto_delete_setter(self):
        self.assertEqual(self.queue_set.auto_delete, True)
        self.assertEqual(self.queue_unset.auto_delete, False)

    def test__specific_setter(self):
        self.assertEqual(self.queue_set.specific_poker, 'xyz')
        self.assertEqual(self.queue_unset.specific_poker, None)
        self.assertEqual(self.queue_set._specific_poker_setter(
            specific_poker=self.PokemanMock(),
            lambda_call=True
        ).__class__, self.PokemanMock().__class__
        )
        self.assertEqual(self.queue_set._specific_poker_setter(
            specific_poker=self.queue_set.specific_poker,
            lambda_call=True
        ), None
        )
        with self.assertRaisesRegex(ValueError, 'Unresolvable specific poker parameter'):
            self.queue_set._specific_poker_setter(
                specific_poker=list(),
                lambda_call=0
            )


if __name__ == '__main__':
    unittest.main()



import sqlite3
import json
import os
from queue import PriorityQueue
import tempfile
from uuid import uuid4

from pokeman.amqp_resources.resolvers import ResourceResolver, Action
from pokeman.amqp_resources.builders import ResourceManager

import logging

LOGGER = logging.getLogger(__name__)


_DB_TYPE = 'file'
_HEAPQ_PREFIX = 'heapq_'
_DB_MODE = 'file'
_DB_CACHE = 'shared'
_MAIN_RESOURCE_TABLE = 'resources'
_POKEMAN_DEFAULT_HEAPQ_ENVIRON = 'POKEMAN_DEFAULT_HEAPQ'


class ResourceHeapQ:
    """
    The ResourceHeapQ managing class, the resources can be managed per process.
    Each Pokeman instance attaches a unique HeapQ to itself, the first to be declared
    will be set default by specifying the process durable environment variable.
    """

    @staticmethod
    def _set_heapq_id(poker_id):
        """
        Static method to set the HeapQ id.

        :param poker_id: The provided poker id.
        :type poker_id: str

        :return: The heapq id.
        :rtype: str
        """
        return '{PREFIX}{POKER_ID}'.format(
            PREFIX=_HEAPQ_PREFIX,
            POKER_ID=poker_id
        )

    @staticmethod
    def _set_heapq_location(heapq_id):
        """
        Static method to set the HeapQ location.

        ..note::
            The location will be set on the Temp
            directory on the Operating System.

        :param heapq_id: The provided heapq id.
        :type heapq_id: str

        :return: The heapq location for the heapq database.
        :rtype: str
        """
        return r'{TEMP_DIR}\{HEAPQ_ID}.heapq'.format(
                TEMP_DIR=tempfile.gettempdir(),
                HEAPQ_ID=heapq_id
            )

    @staticmethod
    def _set_heapq_environment_variable(heapq_id):
        """
        Static method to set the POKEMAN_DEFAULT_HEAPQ environment
        variable for the runtime.

        :param heapq_id: The provided heapq id.
        :type heapq_id: str

        :return: The default status for the provided heapq id.
        :rtype: bool
        """
        if _POKEMAN_DEFAULT_HEAPQ_ENVIRON not in os.environ:
            default_heapq = True
            os.environ[_POKEMAN_DEFAULT_HEAPQ_ENVIRON] = heapq_id
        else:
            default_heapq = False
        return default_heapq

    @staticmethod
    def _heapq_connect(heapq_location):
        """
        Static method to connect to the HeapQ database.

        :param heapq_location: The provided heapq location.
        :type heapq_location: str

        :return: The database connection handler.
        :rtype: sqlite3.Connection
        """
        _connection = sqlite3.connect('{DB_TYPE}:{LOCATION}?cache={CACHE}'.format(
            DB_TYPE=_DB_TYPE,
            LOCATION=heapq_location,
            CACHE=_DB_CACHE
        ), uri=True)
        return _connection

    @classmethod
    def create_database(cls, poker_id):
        """
        Method to create the poker specific HeapQ database in the
        operating system temporary directory. The first poker that is
        instantiated in the current process will set it's HeapQ as default.

        :param poker_id: The provided poker id.
        :type poker_id: str
        """
        heapq_id = cls._set_heapq_id(poker_id=poker_id)
        default_heapq = cls._set_heapq_environment_variable(heapq_id=heapq_id)
        LOGGER.debug('Creating HeapQ database. Default HeapQ: {DEFAULT}'.format(DEFAULT=default_heapq))
        heapq_location = cls._set_heapq_location(heapq_id=heapq_id)
        heapq_connection = cls._heapq_connect(heapq_location=heapq_location)
        cursor = heapq_connection.cursor()
        cursor.execute('''
            CREATE TABLE IF NOT EXISTS {TABLE}(
            id TEXT PRIMARY KEY, 
            priority TEXT, 
            blueprint TEXT,
            poker_id TEXT,
            status TEXT
            )
        '''.format(TABLE=_MAIN_RESOURCE_TABLE))
        LOGGER.debug('Creating HeapQ database. Default HeapQ: {DEFAULT} OK!'.format(DEFAULT=default_heapq))

    @classmethod
    def add_resource(cls, resource, specific_poker=None):
        """
        This method is invoked by an AMQP resource, passing itself
        as function argument. It can also provide a specific poker
        Pokeman instance to only let that specific poker apply the
        resources to the AMQP broker, while the apply_resources method
        also applies the pending resources in default HeapQ.

        :param resource: The provided resource.
        :type resource: Exchange, Queue

        :param specific_poker: The provided specific poker id.
        :type specific_poker: str
        """
        LOGGER.debug('Adding resource to HeapQ: {RESOURCE}. Specific poker: {POKER_ID}'.format(
            RESOURCE=resource.__class__.__name__,
            POKER_ID=str(specific_poker)
        )
        )
        if specific_poker is None and _POKEMAN_DEFAULT_HEAPQ_ENVIRON in os.environ:
            heapq_save_id = os.environ[_POKEMAN_DEFAULT_HEAPQ_ENVIRON]
            heapq_id = None
        elif specific_poker is not None:
            heapq_save_id = cls._set_heapq_id(poker_id=specific_poker)
            heapq_id = specific_poker
        else:
            raise SyntaxError('No default resource heapq found. Make sure to initialize the '
                              'pokeman first before declaring resources.')
        heapq_db_location = r'{TEMP_DIR}\{HEAPQ_ID}.heapq'.format(
                TEMP_DIR=tempfile.gettempdir(),
                HEAPQ_ID=heapq_save_id
            )
        _resource_database = sqlite3.connect('{DB_TYPE}:{LOCATION}?&cache={CACHE}'.format(
            DB_TYPE=_DB_TYPE,
            LOCATION=heapq_db_location,
            CACHE=_DB_CACHE
        ), uri=True)
        resource_resolver = ResourceResolver()
        template = resource_resolver.resolve(resource=resource, action=Action.CREATE)
        str_template = json.dumps(template)
        priority = template['priority']
        new_status = "PENDING"
        resource_id = str(uuid4())
        cursor = _resource_database.cursor()
        cursor.execute('''INSERT INTO ''' + _MAIN_RESOURCE_TABLE + '''(id, priority, blueprint, poker_id, status)
                  VALUES(?,?,?,?,?)''',
                       (resource_id, priority, str_template, heapq_id, new_status))
        _resource_database.commit()
        LOGGER.debug('Adding resource to HeapQ: {RESOURCE} OK!'.format(RESOURCE=resource.__class__.__name__))

    @classmethod
    def fetch_resources(cls, poker_id):
        """
        This method fetches the poker specific pending resources from
        the HeapQ database and also fetches the pending resources from
        the default HeapQ databse.

        :param poker_id: The provided poker id.
        :type poker_id: str

        :return: The list of pending resources.
        :rtype: list
        """
        LOGGER.debug('Fetching resources from HeapQ')
        if _POKEMAN_DEFAULT_HEAPQ_ENVIRON not in os.environ:
            raise SyntaxError('No default resource heapq found. Make sure to initialize the '
                              'pokeman first before declaring resources.')
        specific_heapq_id = {
            'heapq_id': cls._set_heapq_id(poker_id=poker_id),
            'default_flag': False
        }
        default_heapq_id = {
            'heapq_id': os.environ[_POKEMAN_DEFAULT_HEAPQ_ENVIRON],
            'default_flag': True
        }
        resources_list = []
        if specific_heapq_id['heapq_id'] == default_heapq_id['heapq_id']:
            id_list = [specific_heapq_id, default_heapq_id]
        else:
            id_list = [specific_heapq_id, default_heapq_id]
        for _id in id_list:
            heapq_db_location = r'{TEMP_DIR}\{HEAPQ_ID}.heapq'.format(
                TEMP_DIR=tempfile.gettempdir(),
                HEAPQ_ID=_id['heapq_id']
            )
            _resource_database = sqlite3.connect('{DB_TYPE}:{LOCATION}?cache={CACHE}'.format(
                DB_TYPE=_DB_TYPE,
                LOCATION=heapq_db_location,
                CACHE=_DB_CACHE
            ), uri=True)
            _id['heapq_id'] = _id['heapq_id'].replace('heapq_', '')
            if _id['default_flag'] is True:
                _id['heapq_id'] = None
            select_status = 'PENDING'
            cursor = _resource_database.cursor()
            cursor.execute('''SELECT id, priority, blueprint, poker_id, status FROM ''' + _MAIN_RESOURCE_TABLE
                           + ''' WHERE poker_id IS ? AND status IS ?''', (_id['heapq_id'], select_status))
            rows = cursor.fetchall()
            for row in rows:
                new_status = 'FETCHED'
                cursor.execute(
                    '''UPDATE ''' + _MAIN_RESOURCE_TABLE + ''' SET status = ? WHERE id = ?''',
                    (new_status, row[0]))
                resources_list.append(row)
        LOGGER.debug('Fetching resources from HeapQ OK!')
        return resources_list

    @classmethod
    def get_heapq(cls, poker_id, reverse=False):
        """
        This method sets the HeapQ, by declaring the pending resources
        list as queue.PriorityQueue.

        :param poker_id: The provided poker id.
        :type poker_id: str

        :param reverse: The provided reversed paramater
        :type reverse: bool

        :return: The HeapQ.
        :rtype: queue.PriorityQueue
        """
        LOGGER.debug('Get the HeapQ')
        resources_heapq = PriorityQueue()
        resource_list = cls.fetch_resources(poker_id=poker_id)
        if reverse is False:
            for _resource in resource_list:
                resources_heapq.put((int(_resource[1]), _resource[2]))
        elif reverse is True:
            for _resource in resource_list:
                resources_heapq.put((-int(_resource[1]), _resource[2]))
        LOGGER.debug('Get the HeapQ OK!')
        return resources_heapq

    # TODO: Add FETCHED > CREATED HeapQ update
    @classmethod
    def apply_resources(cls, connection, poker_id):
        """
        This method applies the resources from the HeapQ to the
        AMQP broker. It uses the ResourceManager to select the
        right resource builder and and to create the resource.

        :param connection: The provided connection.
        :type connection: pokeman.Pokeman.connection

        :param poker_id: The provided poker id.
        :type poker_id: str

        .. note::
            The ResourceManager handles the actual creation
            of the resources by using the provided connection.
        """
        LOGGER.debug('Pokeman {POKER_ID} applying resources'.format(POKER_ID=poker_id))
        resource_manager = ResourceManager(connection=connection)
        _heapq = cls.get_heapq(poker_id=poker_id)
        for _ in range(0, _heapq.qsize()):
            heap_record = _heapq.get()
            resource_blueprint = json.loads(heap_record[1])
            resource_manager.pick_builder(blueprint=resource_blueprint)
            resource_manager.create_resource()
        LOGGER.debug('Pokeman {POKER_ID} applying resources OK!'.format(POKER_ID=poker_id))

    @classmethod
    def remove_heapq(cls, poker_id):
        """
        This method removes the HeapQ database for the provided Pokeman id.

        :param poker_id: The provided Pokeman id.
        :type poker_id: str
        """
        LOGGER.debug('Removing HeapQ')
        heapq_id = cls._set_heapq_id(poker_id=poker_id)
        heapq_db_location = r'{TEMP_DIR}\{HEAPQ_ID}.heapq'.format(
            TEMP_DIR=tempfile.gettempdir(),
            HEAPQ_ID=heapq_id
        )
        os.remove(heapq_db_location)
        LOGGER.debug('Removing HeapQ OK!')


import unittest
from unittest import mock

from tests import TestAttributes as TA

from pokeman.amqp_resources import resolvers

class ActionTests(unittest.TestCase):
    def setUp(self):
        self.Action = resolvers.Action

    def test_base_name(self):
        self.assertEqual(self.Action.__name__, 'Action')

    def test_enums(self):
        self.assertEqual(self.Action.CREATE.name, 'CREATE')
        self.assertEqual(self.Action.DELETE.name, 'DELETE')


class PriorityTests(unittest.TestCase):
    def setUp(self):
        self.Priority = resolvers.Priority

    def test_base_name(self):
        self.assertEqual(self.Priority.__name__, 'Priority')

    def test_flags(self):
        self.assertEqual(self.Priority.EXCHANGE.name, 'EXCHANGE')
        self.assertEqual(self.Priority.QUEUE.name, 'QUEUE')


class AbstractResourceResolverTests(unittest.TestCase):
    def test_base_name(self):
        self.assertEqual(resolvers.AbstractResourceResolver.__name__, 'AbstractResourceResolver')

    @mock.patch.object(resolvers.AbstractResourceResolver, '__abstractmethods__', set())
    def test_base_initialization(self):
        self.abc_resolver = resolvers.AbstractResourceResolver(
            resource=None,
            action=resolvers.Action.CREATE
        )
        blueprint_fields = ['type', 'priority', 'context', 'action']
        assert [field in self.abc_resolver.blueprint.keys() for field in blueprint_fields]


class ExchangeResolverTests(unittest.TestCase):
    def setUp(self):
        from pokeman.amqp_resources.heapq import ResourceHeapQ
        from pokeman.amqp_resources.globals import Exchange
        with TA.patch(ResourceHeapQ, 'add_resource', lambda resource, specific_poker: None):
            self.ExchangeResolver = resolvers.ExchangeResolver(
                resource=Exchange(),
                action=resolvers.Action.CREATE
            )

    def test_base_name(self):
        self.assertEqual(self.ExchangeResolver.__class__.__name__, 'ExchangeResolver')

    def test_resolver(self):
        from pokeman.amqp_resources.globals import Exchange
        context = {
            'exchange_name': 'pokeman_exchange',
            'exchange_type': 'direct',
            'durable': True,
            'auto_delete': False,
            'specific_poker': None
        }
        self.assertTrue(isinstance(self.ExchangeResolver.resource, Exchange))
        self.assertEqual(self.ExchangeResolver.blueprint['type'], 'Exchange')
        self.assertEqual(self.ExchangeResolver.blueprint['priority'], resolvers.Priority.EXCHANGE)
        self.assertEqual(self.ExchangeResolver.blueprint['action'], resolvers.Action.CREATE.name)
        self.assertEqual(self.ExchangeResolver.blueprint['context'], context)


class QueueResolverTests(unittest.TestCase):
    def setUp(self):
        from pokeman.amqp_resources.heapq import ResourceHeapQ
        from pokeman.amqp_resources.globals import Queue
        with TA.patch(ResourceHeapQ, 'add_resource', lambda resource, specific_poker: None):
            self.QueueResolver = resolvers.QueueResolver(
                resource=Queue(),
                action=resolvers.Action.CREATE
            )

    def test_base_name(self):
        self.assertEqual(self.QueueResolver.__class__.__name__, 'QueueResolver')

    def test_resolver(self):
        from pokeman.amqp_resources.globals import Queue
        context = {
            'queue_name': 'pokeman_queue',
            'exchange': 'pokeman_exchange',
            'routing_key': None,
            'exclusive': False,
            'durable': True,
            'auto_delete': False,
            'specific_poker': None
        }
        self.assertTrue(isinstance(self.QueueResolver.resource, Queue))
        self.assertEqual(self.QueueResolver.blueprint['type'], 'Queue')
        self.assertEqual(self.QueueResolver.blueprint['priority'], resolvers.Priority.QUEUE)
        self.assertEqual(self.QueueResolver.blueprint['action'], resolvers.Action.CREATE.name)
        self.assertEqual(self.QueueResolver.blueprint['context'], context)


class ResourceResolverTests(unittest.TestCase):
    class InvalidResourceMock:
        pass

    def setUp(self):
        self.ResourceResolver = resolvers.ResourceResolver()

    def test_base_name(self):
        self.assertEqual(self.ResourceResolver.__class__.__name__, 'ResourceResolver')

    def test_base_initialization(self):
        self.assertEqual(self.ResourceResolver._blueprint, None)

    def test__template_mapping(self):
        for resource, mapping in self.ResourceResolver._templates.items():
            assert resource.__name__ == mapping['blueprint'].__name__.replace('Resolver', '')

    def test_resolve(self):
        from pokeman.amqp_resources.heapq import ResourceHeapQ
        from pokeman.amqp_resources.globals import Queue
        with TA.patch(ResourceHeapQ, 'add_resource', lambda resource, specific_poker: None):
            blueprint = {
            'type': 'Queue',
            'priority': 2,
            'context': {
                'auto_delete': False,
                'durable': True,
                'exchange': 'pokeman_exchange',
                'exclusive': False,
                'queue_name': 'pokeman_queue',
                'routing_key': None,
                'specific_poker': None
            },
            'action': 'CREATE'
            }
            self.assertEqual(self.ResourceResolver.resolve(
                resource=Queue(),
                action=resolvers.Action.CREATE
            ), blueprint)

    def test_blueprint_valid_resource(self):
        from pokeman.amqp_resources.heapq import ResourceHeapQ
        from pokeman.amqp_resources.globals import Exchange
        with TA.patch(ResourceHeapQ, 'add_resource', lambda resource, specific_poker: None):
            blueprint = {
            'type': 'Exchange',
            'priority': 1,
            'context': {
                'exchange_name': 'pokeman_exchange',
                'exchange_type': 'direct',
                'durable': True,
                'auto_delete': False,
                'specific_poker': None
            },
            'action': 'CREATE'
            }
            self.assertEqual(self.ResourceResolver.blueprint(
                resource=Exchange(),
                action=resolvers.Action.CREATE
            ), blueprint)

    def test_blueprint_invalid_resource(self):
        with self.assertRaisesRegex(ValueError, 'No resource resolver implemented'):
            self.ResourceResolver.blueprint(
                resource=self.InvalidResourceMock(),
                action=resolvers.Action.CREATE
            )



if __name__ == '__main__':
    unittest.main()



import re
import sys
from setuptools import setup, find_packages

with open('pokeman/__init__.py', 'r') as f:
    version = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                        f.read(), re.MULTILINE).group(1)

with open('README.rst', 'r') as f:
    long_description = f.read()

required = [
    'pika'
]

setup(
    name='pokeman',
    version=version,
    description='Pokeman RabbitMQ Service Choreography Library',
    long_description=long_description,
    maintainer='Wolf Marcuse',
    maintainer_email='wmarcuse@gmail.com',
    url='https://github.com/wmarcuse/pokeman',
    packages=find_packages('.'),
    license='BSD',
    install_requires=required,
    package_data={'': ['LICENSE', 'README.rst']},
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Communications', 'Topic :: Internet',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Networking'
    ],
    zip_safe=True)


import unittest
import abc

from pokeman.utils import custom_abc

class DummyAttributeTests(unittest.TestCase):
    def test_base_name(self):
        self.assertEqual(custom_abc.DummyAttribute.__name__, 'DummyAttribute')


class AbstractAttributeTests(unittest.TestCase):
    def test_abstract_attribute(self):
        self.assertIsInstance(custom_abc.abstract_attribute(), custom_abc.DummyAttribute)
        self.assertEqual(custom_abc.abstract_attribute().__is_abstract_attribute__, True)


class ABCMetaTests(unittest.TestCase):
    class AbstractClass(metaclass=custom_abc.ABCMeta):
        @custom_abc.abstract_attribute
        def test_attribute(self):
            pass

    class AbstractImplementation(AbstractClass):
        def __init__(self, attr=False):
            if attr is True:
                self.test_attribute = True
            else:
                pass

    def test_base_name(self):
        self.assertEqual(custom_abc.ABCMeta.__name__, 'ABCMeta')

    def test_equality(self):
        self.assertTrue(issubclass(custom_abc.ABCMeta, abc.ABCMeta))

    def test_metaclass_abstract_attribute_success(self):
        self.assertEqual(self.AbstractImplementation(attr=True).test_attribute, True)

    def test_metaclass_abstract_attribute_failure(self):
        with self.assertRaisesRegex(NotImplementedError, "Can't instantiate abstract class"):
            self.AbstractImplementation(attr=False)

if __name__ == '__main__':
    unittest.main()


import unittest
from unittest import mock

from tests import TestAttributes as TA

from pokeman.amqp_resources import builders


class AbstractResourceBuilderTests(unittest.TestCase):
    def test_base_name(self):
        self.assertEqual(builders.AbstractResourceBuilder.__name__, 'AbstractResourceBuilder')

    @mock.patch.object(builders.AbstractResourceBuilder, '__abstractmethods__', set())
    def test_base_initialization(self):
        self.abc_builder = builders.AbstractResourceBuilder()
        assert self.abc_builder.digest_blueprint
        assert self.abc_builder.create_resource
        self.assertIsInstance(self.abc_builder, builders.AbstractResourceBuilder)


class AbstractResourceDestroyerTests(unittest.TestCase):
    def test_base_name(self):
        self.assertEqual(builders.AbstractResourceDestroyer.__name__, 'AbstractResourceDestroyer')

    @mock.patch.object(builders.AbstractResourceDestroyer, '__abstractmethods__', set())
    def test_base_initialization(self):
        blueprint = {
            'context': None
        }
        self.abc_destroyer = builders.AbstractResourceDestroyer(
            channel=None,
            blueprint=blueprint
        )
        self.assertEqual(self.abc_destroyer.channel, None)
        self.assertEqual(self.abc_destroyer.blueprint, blueprint)
        self.assertEqual(self.abc_destroyer.blueprint_context, blueprint['context'])
        assert self.abc_destroyer.delete_resource
        self.assertIsInstance(self.abc_destroyer, builders.AbstractResourceDestroyer)


class AbstractExistenceChecker(unittest.TestCase):
    def test_base_name(self):
        self.assertEqual(builders.AbstractExistenceChecker.__name__, 'AbstractExistenceChecker')

    @mock.patch.object(builders.AbstractExistenceChecker, '__abstractmethods__', set())
    def test_base_initialization(self):
        blueprint = {
            'context': None
        }
        self.abc_existence_checker = builders.AbstractExistenceChecker(
            channel=None,
            blueprint=blueprint
        )
        self.assertEqual(self.abc_existence_checker.channel, None)
        self.assertEqual(self.abc_existence_checker.blueprint, blueprint)
        self.assertEqual(self.abc_existence_checker.blueprint_context, blueprint['context'])
        self.assertEqual(self.abc_existence_checker._exists, False)
        self.assertIsInstance(self.abc_existence_checker, builders.AbstractExistenceChecker)
        assert not self.abc_existence_checker.exists
        assert self.abc_existence_checker.check_existence

    @mock.patch.object(builders.AbstractExistenceChecker, '__abstractmethods__', set())
    def test_reset(self):
        blueprint = {
            'context': None
        }
        self.abc_existence_checker = builders.AbstractExistenceChecker(
            channel=None,
            blueprint=blueprint
        )
        self.abc_existence_checker._exists = True
        self.abc_existence_checker.reset()
        self.assertEqual(self.abc_existence_checker._exists, False)

    @mock.patch.object(builders.AbstractExistenceChecker, '__abstractmethods__', set())
    def test_exists(self):
        blueprint = {
            'context': None
        }
        self.abc_existence_checker = builders.AbstractExistenceChecker(
            channel=None,
            blueprint=blueprint
        )
        self.abc_existence_checker._exists = True
        self.assertEqual(self.abc_existence_checker.exists, True)
        self.assertEqual(self.abc_existence_checker._exists, False)


class ExchangeBuilderTests(unittest.TestCase):
    class ChannelMock:
        @classmethod
        def exchange_declare(cls, exchange, exchange_type, durable, auto_delete):
            pass

    def setUp(self):
        self.blueprint = {
            'type': 'Exchange',
            'priority': 1,
            'context': {
                'exchange_name': 'pokeman_exchange',
                'exchange_type': 'direct',
                'durable': True,
                'auto_delete': False,
                'specific_poker': None
            },
            'action': 'CREATE'
        }
        self.blueprint_context = self.blueprint['context']
        self.ExchangeBuilder = builders.ExchangeBuilder(
            channel=self.ChannelMock(),
            blueprint=self.blueprint
        )

    def test_base_name(self):
        self.assertEqual(self.ExchangeBuilder.__class__.__name__, 'ExchangeBuilder')

    def test_digest_blueprint_success(self):
        self.ExchangeBuilder.digest_blueprint()
        self.assertEqual(self.ExchangeBuilder.exchange_name, self.blueprint_context['exchange_name'])
        self.assertEqual(self.ExchangeBuilder.exchange_type, self.blueprint_context['exchange_type'])
        self.assertEqual(self.ExchangeBuilder.durable, self.blueprint_context['durable'])
        self.assertEqual(self.ExchangeBuilder.auto_delete, self.blueprint_context['auto_delete'])
        self.assertEqual(self.ExchangeBuilder.specific_poker, self.blueprint_context['specific_poker'])

    def test_digest_blueprint_failure(self):
        mock_blueprint = self.blueprint
        mock_blueprint['type'] = 'Undefined'
        with self.assertRaisesRegex(ValueError, 'received the wrong blueprint'):
            self.ExchangeBuilder.blueprint = mock_blueprint
            self.ExchangeBuilder.digest_blueprint()

    def test_reset(self):
        self.ExchangeBuilder.digest_blueprint()
        self.ExchangeBuilder.reset()
        attributes = [
            self.ExchangeBuilder.exchange_name,
            self.ExchangeBuilder.exchange_type,
            self.ExchangeBuilder.durable,
            self.ExchangeBuilder.auto_delete,
            self.ExchangeBuilder.specific_poker
        ]
        for attr in attributes:
            self.assertEqual(attr, None)

    def test_create_resource(self):
        self.assertEqual(self.ExchangeBuilder.create_resource(), None)


class QueueBuilderTests(unittest.TestCase):
    class ChannelMock:
        @classmethod
        def queue_declare(cls, queue, durable, exclusive, auto_delete):
            pass

        @classmethod
        def queue_bind(cls, queue, exchange, routing_key):
            pass

    def setUp(self):
        self.blueprint = {
            'type': 'Queue',
            'priority': 2,
            'context': {
                'auto_delete': False,
                'durable': True,
                'exchange': 'pokeman_exchange',
                'exclusive': False,
                'queue_name': 'pokeman_queue',
                'routing_key': None,
                'specific_poker': None
            },
            'action': 'CREATE'
        }
        self.blueprint_context = self.blueprint['context']
        self.QueueBuilder = builders.QueueBuilder(
            channel=self.ChannelMock(),
            blueprint=self.blueprint
        )

    def test_base_name(self):
        self.assertEqual(self.QueueBuilder.__class__.__name__, 'QueueBuilder')

    def test_digest_blueprint_success(self):
        self.QueueBuilder.digest_blueprint()
        self.assertEqual(self.QueueBuilder.queue_name, self.blueprint_context['queue_name'])
        self.assertEqual(self.QueueBuilder.exchange, self.blueprint_context['exchange'])
        self.assertEqual(self.QueueBuilder.routing_key, self.blueprint_context['routing_key'])
        self.assertEqual(self.QueueBuilder.exclusive, self.blueprint_context['exclusive'])
        self.assertEqual(self.QueueBuilder.durable, self.blueprint_context['durable'])
        self.assertEqual(self.QueueBuilder.auto_delete, self.blueprint_context['auto_delete'])
        self.assertEqual(self.QueueBuilder.specific_poker, self.blueprint_context['specific_poker'])

    def test_digest_blueprint_failure(self):
        mock_blueprint = self.blueprint
        mock_blueprint['type'] = 'Undefined'
        with self.assertRaisesRegex(ValueError, 'received the wrong blueprint'):
            self.QueueBuilder.blueprint = mock_blueprint
            self.QueueBuilder.digest_blueprint()

    def test_reset(self):
        self.QueueBuilder.digest_blueprint()
        self.QueueBuilder.reset()
        attributes = [
            self.QueueBuilder.queue_name,
            self.QueueBuilder.exchange,
            self.QueueBuilder.routing_key,
            self.QueueBuilder.exclusive,
            self.QueueBuilder.durable,
            self.QueueBuilder.auto_delete,
            self.QueueBuilder.specific_poker
        ]
        for attr in attributes:
            self.assertEqual(attr, None)

    def test_create_resource(self):
        self.assertEqual(self.QueueBuilder.create_resource(), None)


class ExchangeExistenceCheckerTests(unittest.TestCase):
    class ChannelMock:
        def __init__(self, exists=False):
            self.exists = exists

        def exchange_declare(self, exchange, exchange_type, passive):
            from pika.exceptions import ChannelClosedByBroker
            if self.exists is True:
                return None
            else:
                raise ChannelClosedByBroker(406, reply_text='... NOT FOUND ...')

    def setUp(self):
        self.blueprint = {
            'type': 'Exchange',
            'priority': 1,
            'context': {
                'exchange_name': 'pokeman_exchange',
                'exchange_type': 'direct',
                'durable': True,
                'auto_delete': False,
                'specific_poker': None
            },
            'action': 'CREATE'
        }
        self.ExchangeExistenceChecker = builders.ExchangeExistenceChecker

    def test_base_name(self):
        self.assertEqual(self.ExchangeExistenceChecker.__name__, 'ExchangeExistenceChecker')

    def test_check_existence_true(self):
        self.ExchangeExistenceChecker = self.ExchangeExistenceChecker(
            channel=self.ChannelMock(
                exists=True
            ),
            blueprint=self.blueprint
        )
        self.ExchangeExistenceChecker.check_existence()
        self.assertEqual(self.ExchangeExistenceChecker._exists, True)

    def test_check_existence_false(self):
        self.ExchangeExistenceChecker = self.ExchangeExistenceChecker(
            channel=self.ChannelMock(
                exists=False
            ),
            blueprint=self.blueprint
        )
        self.ExchangeExistenceChecker.check_existence()
        self.assertEqual(self.ExchangeExistenceChecker._exists, False)


class QueueExistenceCheckerTests(unittest.TestCase):
    class ChannelMock:
        def __init__(self, exists=False):
            self.exists = exists

        def queue_declare(self, queue, passive):
            from pika.exceptions import ChannelClosedByBroker
            if self.exists is True:
                return None
            else:
                raise ChannelClosedByBroker(406, reply_text='... NOT FOUND ...')

    def setUp(self):
        self.blueprint = {
            'type': 'Queue',
            'priority': 2,
            'context': {
                'auto_delete': False,
                'durable': True,
                'exchange': 'pokeman_exchange',
                'exclusive': False,
                'queue_name': 'pokeman_queue',
                'routing_key': None,
                'specific_poker': None
            },
            'action': 'CREATE'
        }
        self.QueueExistenceChecker = builders.QueueExistenceChecker

    def test_base_name(self):
        self.assertEqual(self.QueueExistenceChecker.__name__, 'QueueExistenceChecker')

    def test_check_existence_true(self):
        self.QueueExistenceChecker = self.QueueExistenceChecker(
            channel=self.ChannelMock(
                exists=True
            ),
            blueprint=self.blueprint
        )
        self.QueueExistenceChecker.check_existence()
        self.assertEqual(self.QueueExistenceChecker._exists, True)

    def test_check_existence_false(self):
        self.QueueExistenceChecker = self.QueueExistenceChecker(
            channel=self.ChannelMock(
                exists=False
            ),
            blueprint=self.blueprint
        )
        self.QueueExistenceChecker.check_existence()
        self.assertEqual(self.QueueExistenceChecker._exists, False)


class ExchangeDestroyerTests(unittest.TestCase):
    class ChannelMock:
        def __init__(self, deletable=False, reply=None):
            self.deletable = deletable
            self.reply = reply

        def exchange_delete(self, exchange):
            from pika.exceptions import ChannelClosedByBroker
            if self.deletable is True:
                return None
            else:
                raise ChannelClosedByBroker(406, reply_text=self.reply)

    def setUp(self):
        self.blueprint = {
            'type': 'Exchange',
            'priority': 1,
            'context': {
                'exchange_name': 'pokeman_exchange',
                'exchange_type': 'direct',
                'durable': True,
                'auto_delete': False,
                'specific_poker': None
            },
            'action': 'CREATE'
        }
        self.ExchangeDestroyer = builders.ExchangeDestroyer

    def test_base_name(self):
        self.assertEqual(self.ExchangeDestroyer.__name__, 'ExchangeDestroyer')

    def test_delete_resource_deletable_true(self):
        self.ExchangeDestroyer = self.ExchangeDestroyer(
            channel=self.ChannelMock(
                deletable=True
            ),
            blueprint=self.blueprint
        )
        self.assertEqual(self.ExchangeDestroyer.delete_resource(), None)

    def test_delete_resource_deletable_false_resource_does_not_exists(self):
        self.ExchangeDestroyer = self.ExchangeDestroyer(channel=self.ChannelMock(
                deletable=False,
                reply='... NOT_FOUND ...'
            ), blueprint=self.blueprint)
        self.ExchangeDestroyer.delete_resource()
        self.assertEqual(self.ExchangeDestroyer.delete_resource(), None)


class QueueDestroyerTests(unittest.TestCase):
    class ChannelMock:
        def __init__(self, deletable=False, reply=None):
            self.deletable = deletable
            self.reply = reply

        def queue_delete(self, queue):
            from pika.exceptions import ChannelClosedByBroker
            if self.deletable is True:
                return None
            else:
                raise ChannelClosedByBroker(406, reply_text=self.reply)

    def setUp(self):
        self.blueprint = {
            'type': 'Queue',
            'priority': 2,
            'context': {
                'auto_delete': False,
                'durable': True,
                'exchange': 'pokeman_exchange',
                'exclusive': False,
                'queue_name': 'pokeman_queue',
                'routing_key': None,
                'specific_poker': None
            },
            'action': 'CREATE'
        }

        self.QueueDestroyer = builders.QueueDestroyer

    def test_base_name(self):
        self.assertEqual(self.QueueDestroyer.__name__, 'QueueDestroyer')

    def test_delete_resource_deletable_true(self):
        self.QueueDestroyer = self.QueueDestroyer(
            channel=self.ChannelMock(
                deletable=True
            ),
            blueprint=self.blueprint
        )
        self.assertEqual(self.QueueDestroyer.delete_resource(), None)

    def test_delete_resource_deletable_false_resource_does_not_exists(self):
        self.QueueDestroyer = self.QueueDestroyer(channel=self.ChannelMock(
                deletable=False,
                reply='... NOT_FOUND ...'
            ), blueprint=self.blueprint)
        self.QueueDestroyer.delete_resource()
        self.assertEqual(self.QueueDestroyer.delete_resource(), None)


class ResourceManagerTests(unittest.TestCase):

    class ConnectionMock:
        def __init__(self, resource, exists=False, scope='existence_check'):
            self._exists = exists
            self.scope = scope

            class Sub:
                @property
                def exists(self):
                    if self._exists is True:
                        return True
                    else:
                        return False

            setattr(Sub, '_exists', self._exists)
            self.channel = Sub
            if resource == 'Exchange':
                setattr(self.channel, 'exchange_declare', ExchangeExistenceCheckerTests.ChannelMock.exchange_declare)
            elif resource == 'Queue':
                self.channel.queue_declare = QueueExistenceCheckerTests.ChannelMock.queue_declare

    class ResourceHeapQMock:
        _HEAPQ = [
            (
                1,
                {
                    "type": "Exchange",
                    "priority": 1,
                    "context": {
                        "exchange_name": "pokeman_exchange",
                        "exchange_type": "direct",
                        "durable": True,
                        "auto_delete": False,
                        "specific_poker": "4a06075e-25b0-49e8-b000-7feca505efb5"
                    },
                    "action": "CREATE"
                }
            ),
            (
                2,
                {
                    "type": "Queue",
                    "priority": 2,
                    "context": {
                        "queue_name": "pokeman_queue",
                        "exchange": "pokeman_exchange",
                        "routing_key": None,
                        "exclusive": False,
                        "durable": False,
                        "auto_delete": False,
                        "specific_poker": None
                    },
                    "action": "CREATE"
                }
            )
        ]

        @classmethod
        def get_heapq(cls, poker_id, reverse):
            import json
            from queue import PriorityQueue
            resource_heapq = PriorityQueue()
            for _resource in cls._HEAPQ:
                resource_heapq.put((_resource[0], json.dumps(_resource[1])))
            return resource_heapq

    def setUp(self):
        self.exchange_blueprint = {
            'type': 'Exchange',
            'priority': 1,
            'context': {
                'exchange_name': 'pokeman_exchange',
                'exchange_type': 'direct',
                'durable': True,
                'auto_delete': False,
                'specific_poker': None
            },
            'action': 'CREATE'
        }
        self.queue_blueprint = {
            'type': 'Queue',
            'priority': 2,
            'context': {
                'auto_delete': False,
                'durable': True,
                'exchange': 'pokeman_exchange',
                'exclusive': False,
                'queue_name': 'pokeman_queue',
                'routing_key': None,
                'specific_poker': None
            },
            'action': 'CREATE'
        }
        self.ExchangeResourceManagerResourceExists = builders.ResourceManager(
            connection=self.ConnectionMock(
                resource=self.exchange_blueprint['type'],
                exists=True,
                scope='existence_check'
            )
        )
        self.ExchangeResourceManagerResourceNotExists = builders.ResourceManager(
            connection=self.ConnectionMock(
                resource=self.exchange_blueprint['type'],
                exists=False,
                scope='existence_check'
            )
        )
        self.QueueResourceManagerResourceExists = builders.ResourceManager(
            connection=self.ConnectionMock(
                resource=self.queue_blueprint['type'],
                exists=True,
                scope='existence_check'
            )
        )
        self.QueueResourceManagerResourceNotExists = builders.ResourceManager(
            connection=self.ConnectionMock(
                resource=self.queue_blueprint['type'],
                exists=False,
                scope='existence_check'
            )
        )

    def test_base_name(self):
        self.assertEqual(self.ExchangeResourceManagerResourceExists.__class__.__name__, 'ResourceManager')

    def test_templates_mapping(self):
        for resource, mapping in self.ExchangeResourceManagerResourceExists._templates.items():
            self.assertIn('builder', mapping.keys())
            self.assertIn('checker', mapping.keys())
            self.assertIn('destroyer', mapping.keys())

    def test_templates_mapped_resources(self):
        from pokeman import Exchange, Queue
        self.assertIn(Exchange.__name__, self.QueueResourceManagerResourceExists._templates.keys())
        self.assertIn(Queue.__name__, self.ExchangeResourceManagerResourceExists._templates.keys())

    def test_initialization(self):
        self.assertEqual(self.ExchangeResourceManagerResourceExists.builder, None)
        self.assertTrue(isinstance(self.ExchangeResourceManagerResourceExists._connection, self.ConnectionMock))
        self.assertEqual(self.ExchangeResourceManagerResourceExists.resource_exists, False)

    def test_pick_builder_failure_already_assigned(self):
        self.ExchangeResourceManagerResourceExists.builder = 'xyz'
        with self.assertRaisesRegex(SyntaxError, 'already picked a builder'):
            self.ExchangeResourceManagerResourceExists.pick_builder(
                blueprint=self.exchange_blueprint
            )

    def test_pick_builder_failure_invalid_resource_type(self):
        self.exchange_blueprint['type'] = 'Invalid'
        with self.assertRaisesRegex(ValueError, 'The provided resource is not valid'):
            self.ExchangeResourceManagerResourceExists.pick_builder(
                blueprint=self.exchange_blueprint
            )

    def test_pick_builder_exchange_success_resource_exists(self):
        self.ExchangeResourceManagerResourceExists.pick_builder(
            blueprint=self.exchange_blueprint
        )
        self.assertEqual(self.ExchangeResourceManagerResourceExists.resource_exists, True)
        self.assertEqual(self.ExchangeResourceManagerResourceExists.builder, None)

    def test_pick_builder_exchange_success_resource_not_exists(self):
        from pokeman.amqp_resources.builders import ExchangeBuilder
        self.ExchangeResourceManagerResourceNotExists.pick_builder(
            blueprint=self.exchange_blueprint
        )
        self.assertEqual(self.ExchangeResourceManagerResourceNotExists.resource_exists, False)
        self.assertTrue(isinstance(self.ExchangeResourceManagerResourceNotExists.builder, ExchangeBuilder))

    def test_pick_builder_queue_success_resource_exists(self):
        self.QueueResourceManagerResourceExists.pick_builder(
            blueprint=self.queue_blueprint
        )
        self.assertEqual(self.QueueResourceManagerResourceExists.resource_exists, True)
        self.assertEqual(self.QueueResourceManagerResourceExists.builder, None)

    def test_pick_builder_queue_success_resource_not_exists(self):
        from pokeman.amqp_resources.builders import QueueBuilder
        self.QueueResourceManagerResourceNotExists.pick_builder(
            blueprint=self.queue_blueprint
        )
        self.assertEqual(self.QueueResourceManagerResourceNotExists.resource_exists, False)
        self.assertTrue(isinstance(self.QueueResourceManagerResourceNotExists.builder, QueueBuilder))

    def test_create_exchange_resource_resource_exists(self):
        exchange_test_builder_object = ExchangeBuilderTests()
        exchange_test_builder_object.setUp()
        exchange_builder = exchange_test_builder_object.ExchangeBuilder
        self.ExchangeResourceManagerResourceExists.builder = exchange_builder
        self.ExchangeResourceManagerResourceExists.resource_exists = True
        self.assertEqual(self.ExchangeResourceManagerResourceExists.resource_exists, True)
        self.ExchangeResourceManagerResourceExists.create_resource()
        self.assertEqual(self.ExchangeResourceManagerResourceExists.builder, None)

    def test_create_exchange_resource_resource_not_exists(self):
        exchange_test_builder_object = ExchangeBuilderTests()
        exchange_test_builder_object.setUp()
        exchange_builder = exchange_test_builder_object.ExchangeBuilder
        self.ExchangeResourceManagerResourceNotExists.builder = exchange_builder
        self.assertEqual(self.ExchangeResourceManagerResourceNotExists.resource_exists, False)
        self.ExchangeResourceManagerResourceNotExists.create_resource()
        self.assertEqual(self.ExchangeResourceManagerResourceNotExists.builder, None)

    def test_create_queue_resource_resource_exists(self):
        queue_test_builder_object = QueueBuilderTests()
        queue_test_builder_object.setUp()
        queue_builder = queue_test_builder_object.QueueBuilder
        self.QueueResourceManagerResourceExists.builder = queue_builder
        self.QueueResourceManagerResourceExists.resource_exists = True
        self.assertEqual(self.QueueResourceManagerResourceExists.resource_exists, True)
        self.QueueResourceManagerResourceExists.create_resource()
        self.assertEqual(self.QueueResourceManagerResourceExists.builder, None)

    def test_create_queue_resource_resource_not_exists(self):
        queue_test_builder_object = QueueBuilderTests()
        queue_test_builder_object.setUp()
        queue_builder = queue_test_builder_object.QueueBuilder
        self.QueueResourceManagerResourceNotExists.builder = queue_builder
        self.assertEqual(self.QueueResourceManagerResourceNotExists.resource_exists, False)
        self.QueueResourceManagerResourceNotExists.create_resource()
        self.assertEqual(self.QueueResourceManagerResourceNotExists.builder, None)

    def test_destroy_builder(self):
        self.ExchangeResourceManagerResourceExists.builder = 'xyz'
        self.ExchangeResourceManagerResourceExists.resource_exists = True
        self.ExchangeResourceManagerResourceExists.destroy_builder()
        self.assertEqual(self.ExchangeResourceManagerResourceExists.builder, None)
        self.assertEqual(self.ExchangeResourceManagerResourceExists.resource_exists, False)

    def test_delete_attached_resources(self):
        from pokeman.amqp_resources import heapq
        template_mock = {
            'Exchange': {
                'destroyer': lambda channel, blueprint: None
            },
            'Queue': {
                'destroyer': lambda channel, blueprint: None
            }
        }
        with TA.patch(heapq, 'ResourceHeapQ', self.ResourceHeapQMock), \
             TA.patch(builders.ResourceManager, '_templates', template_mock):

            self.assertEqual(self.ExchangeResourceManagerResourceExists.delete_attached_resources(
                poker_id='xyz'
            ), None)



import unittest

from pokeman import BasicConfig

class BasicConfigTests(unittest.TestCase):
    def setUp(self):
        self.basic_config = BasicConfig(
            connection_attempts=5,
            heartbeat=7200,
            retry_delay=2
        )

    def test_base_name(self):
        self.assertEqual(self.basic_config.__class__.__name__, 'BasicConfig')

    def test_base_initialization(self):
        self.assertEqual(self.basic_config.CONNECTION_ATTEMPTS, 5)
        self.assertEqual(self.basic_config.HEARTBEAT, 7200)
        self.assertEqual(self.basic_config.RETRY_DELAY, 2)

if __name__ == '__main__':
    unittest.main()
